# Copyright (c) 2011-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    TextIO,
    BinaryIO,
    Iterable,
    Iterator,
    Union,
    Callable,
    cast,
    Optional,
    Sequence,
    Any,
)
import abc
import base64
import io
import logging
import os
import pathlib
from datetime import datetime, timezone
from itertools import chain

import ezdxf
from ezdxf.audit import Auditor
from ezdxf.entities.dxfgroups import GroupCollection
from ezdxf.entities.material import MaterialCollection
from ezdxf.entities.mleader import MLeaderStyleCollection
from ezdxf.entities.mline import MLineStyleCollection
from ezdxf.entitydb import EntityDB
from ezdxf.groupby import groupby
from ezdxf.layouts import Modelspace, Paperspace
from ezdxf.layouts.layouts import Layouts
from ezdxf.lldxf import const
from ezdxf.lldxf.const import (
    BLK_XREF,
    BLK_EXTERNAL,
    DXF13,
    DXF14,
    DXF2000,
    DXF2007,
    DXF12,
    DXF2013,
)
from ezdxf.lldxf import loader
from ezdxf.lldxf.tagwriter import (
    AbstractTagWriter,
    TagWriter,
    BinaryTagWriter,
    JSONTagWriter,
)
from ezdxf.query import EntityQuery
from ezdxf.render.dimension import DimensionRenderer
from ezdxf.sections.acdsdata import AcDsDataSection, new_acds_data_section
from ezdxf.sections.blocks import BlocksSection
from ezdxf.sections.classes import ClassesSection
from ezdxf.sections.entities import EntitySection, StoredSection
from ezdxf.sections.header import HeaderSection
from ezdxf.sections.objects import ObjectsSection
from ezdxf.sections.tables import TablesSection
from ezdxf.tools import guid
from ezdxf.tools.codepage import tocodepage, toencoding
from ezdxf.tools.juliandate import juliandate
from ezdxf.tools.text import safe_string, MAX_STR_LEN
from ezdxf import messenger, msgtypes


logger = logging.getLogger("ezdxf")

if TYPE_CHECKING:
    from ezdxf.entities import DXFEntity, Layer, VPort, Dictionary
    from ezdxf.eztypes import GenericLayoutType
    from ezdxf.layouts import Layout
    from ezdxf.lldxf.tags import Tags
    from ezdxf.lldxf.types import DXFTag
    from ezdxf.sections.tables import (
        LayerTable,
        LinetypeTable,
        TextstyleTable,
        DimStyleTable,
        AppIDTable,
        UCSTable,
        ViewTable,
        ViewportTable,
        BlockRecordTable,
    )

CONST_GUID = "{00000000-0000-0000-0000-000000000000}"
CONST_MARKER_STRING = "0.0 @ 2000-01-01T00:00:00.000000+00:00"
CREATED_BY_EZDXF = "CREATED_BY_EZDXF"
WRITTEN_BY_EZDXF = "WRITTEN_BY_EZDXF"
EZDXF_META = "EZDXF_META"


def _validate_handle_seed(seed: str) -> str:
    from ezdxf.tools.handle import START_HANDLE

    if seed is None:
        seed = START_HANDLE
    try:
        v = int(seed, 16)
        if v < 1:
            seed = START_HANDLE
    except ValueError:
        seed = START_HANDLE
    return seed


class Drawing:
    def __init__(self, dxfversion=DXF2013) -> None:
        self.entitydb = EntityDB()
        self.messenger = messenger.Messenger(self)
        target_dxfversion = dxfversion.upper()
        self._dxfversion: str = const.acad_release_to_dxf_version.get(
            target_dxfversion, target_dxfversion
        )
        if self._dxfversion not in const.versions_supported_by_new:
            raise const.DXFVersionError(f'Unsupported DXF version "{self.dxfversion}".')
        # Store original dxf version if loaded (and maybe converted R13/14)
        # from file.
        self._loaded_dxfversion: Optional[str] = None

        # Status flag which is True while loading content from a DXF file:
        self.is_loading = False
        self.encoding: str = "cp1252"  # read/write
        self.filename: Optional[str] = None

        # Reason for using "type: ignore".
        # I won't use "Optional" for the following attributes because these
        # objects are required, just not yet! Setting up an empty DXF Document
        # is not necessary if the document is read from the file system,
        # see class-methods new() and read().

        # named objects dictionary
        self.rootdict: Dictionary = None  # type: ignore

        # DXF sections
        self.header: HeaderSection = None  # type: ignore
        self.classes: ClassesSection = None  # type: ignore
        self.tables: TablesSection = None  # type: ignore
        self.blocks: BlocksSection = None  # type: ignore
        self.entities: EntitySection = None  # type: ignore
        self.objects: ObjectsSection = None  # type: ignore

        # DXF R2013 and later
        self.acdsdata: AcDsDataSection = None  # type: ignore

        self.stored_sections: list[StoredSection] = []
        self.layouts: Layouts = None  # type: ignore
        self.groups: GroupCollection = None  # type: ignore
        self.materials: MaterialCollection = None  # type: ignore
        self.mleader_styles: MLeaderStyleCollection = None  # type: ignore
        self.mline_styles: MLineStyleCollection = None  # type: ignore

        # Set to False if the generated DXF file will be incompatible to AutoCAD
        self._acad_compatible = True
        # Store reasons for AutoCAD incompatibility:
        self._acad_incompatibility_reason: set[str] = set()

        # DIMENSION rendering engine can be replaced by a custom Dimension
        # render: see property Drawing.dimension_renderer
        self._dimension_renderer = DimensionRenderer()

        # Some fixes can't be applied while the DXF document is not fully
        # initialized, store this fixes as callable object:
        self._post_init_commands: list[Callable] = []
        # Don't create any new entities here:
        # New created handles could collide with handles loaded from DXF file.
        assert len(self.entitydb) == 0

    @classmethod
    def new(cls, dxfversion: str = DXF2013) -> Drawing:
        """Create new drawing. Package users should use the factory function
        :func:`ezdxf.new`. (internal API)
        """
        doc = cls(dxfversion)
        doc._setup()
        doc._create_ezdxf_metadata()
        return doc

    def _setup(self):
        self.header = HeaderSection.new()
        self.classes = ClassesSection(self)
        self.tables = TablesSection(self)
        self.blocks = BlocksSection(self)
        self.entities = EntitySection(self)
        self.objects = ObjectsSection(self)
        self.acdsdata = new_acds_data_section(self)
        self.rootdict = self.objects.rootdict
        # Create missing tables:
        self.objects.setup_object_management_tables(self.rootdict)
        self.layouts = Layouts.setup(self)
        self._finalize_setup()

    def _finalize_setup(self):
        """Common setup tasks for new and loaded DXF drawings."""
        self.groups = GroupCollection(self)
        self.materials = MaterialCollection(self)

        self.mline_styles = MLineStyleCollection(self)
        # all required internal structures are ready
        # now do the stuff to please AutoCAD
        self._create_required_table_entries()

        # mleader_styles requires text styles
        self.mleader_styles = MLeaderStyleCollection(self)
        self._set_required_layer_attributes()
        self._setup_metadata()
        self._execute_post_init_commands()

    def _execute_post_init_commands(self):
        for cmd in self._post_init_commands:
            cmd()
        del self._post_init_commands

    def _create_required_table_entries(self):
        self._create_required_vports()
        self._create_required_linetypes()
        self._create_required_layers()
        self._create_required_styles()
        self._create_required_appids()
        self._create_required_dimstyles()

    def _set_required_layer_attributes(self):
        for layer in self.layers:
            layer.set_required_attributes()

    def _create_required_vports(self):
        if "*Active" not in self.viewports:
            self.viewports.new("*Active")

    def _create_required_appids(self):
        if "ACAD" not in self.appids:
            self.appids.new("ACAD")

    def _create_required_linetypes(self):
        linetypes = self.linetypes
        for name in ("ByBlock", "ByLayer", "Continuous"):
            if name not in linetypes:
                linetypes.new(name)

    def _create_required_dimstyles(self):
        if "Standard" not in self.dimstyles:
            self.dimstyles.new("Standard")

    def _create_required_styles(self):
        if "Standard" not in self.styles:
            self.styles.new("Standard")

    def _create_required_layers(self):
        layers = self.layers
        if "0" not in layers:
            layers.new("0")
        if "Defpoints" not in layers:
            layers.new("Defpoints", dxfattribs={"plot": 0})  # do not plot
        else:
            # AutoCAD requires a plot flag = 0
            layers.get("Defpoints").dxf.plot = 0

    def _setup_metadata(self):
        self.header["$ACADVER"] = self.dxfversion
        self.header["$TDCREATE"] = juliandate(datetime.now())
        if self.header.get("$FINGERPRINTGUID", CONST_GUID) == CONST_GUID:
            self.reset_fingerprint_guid()
        if self.header.get("$VERSIONGUID", CONST_GUID) == CONST_GUID:
            self.reset_version_guid()

    @property
    def dxfversion(self) -> str:
        """Get current DXF version."""
        return self._dxfversion

    @dxfversion.setter
    def dxfversion(self, version: str) -> None:
        """Set current DXF version."""
        self._dxfversion = self._validate_dxf_version(version)
        self.header["$ACADVER"] = version

    @property
    def loaded_dxfversion(self) -> Optional[str]:
        return self._loaded_dxfversion

    @property
    def output_encoding(self):
        """Returns required output encoding for writing document to a text
        streams.
        """
        return "utf-8" if self.dxfversion >= DXF2007 else self.encoding

    @property
    def units(self) -> int:
        """Get and set the document/modelspace base units as enum, for more
        information read this: :ref:`dxf units`. Requires DXF R2000 or newer.

        """
        return self.header.get("$INSUNITS", 0)

    @units.setter
    def units(self, unit_enum: int) -> None:
        if 0 <= unit_enum < 25:
            if self.dxfversion < DXF2000:
                logger.warning(
                    "Drawing units ($INSUNITS) are not exported for DXF R12."
                )
            self.header["$INSUNITS"] = unit_enum
        else:
            raise ValueError(f"Invalid units enum: {unit_enum}")

    def _validate_dxf_version(self, version: str) -> str:
        version = version.upper()
        # translates 'R12' -> 'AC1009'
        version = const.acad_release_to_dxf_version.get(version, version)
        if version not in const.versions_supported_by_save:
            raise const.DXFVersionError(f'Unsupported DXF version "{version}".')
        if version == DXF12:
            if self._dxfversion > DXF12:
                logger.warning(
                    f"Downgrade from DXF {self.acad_release} to R12 may create "
                    f"an invalid DXF file."
                )
        elif version < self._dxfversion:
            logger.info(
                f"Downgrade from DXF {self.acad_release} to "
                f"{const.acad_release[version]} can cause lost of features."
            )
        return version

    def get_abs_filepath(self) -> pathlib.Path:
        """Returns the absolute filepath of the document."""
        if not self.filename:
            return pathlib.Path(".")
        return pathlib.Path(self.filename).resolve()

    @classmethod
    def read(cls, stream: TextIO) -> Drawing:
        """Open an existing drawing. Package users should use the factory
        function :func:`ezdxf.read`. To preserve possible binary data in
        XRECORD entities use :code:`errors='surrogateescape'` as error handler
        for the import stream.

        Args:
             stream: text stream yielding text (unicode) strings by readline()

        """
        from .lldxf.tagger import ascii_tags_loader

        tag_loader = ascii_tags_loader(stream)
        return cls.load(tag_loader)

    @classmethod
    def load(cls, tag_loader: Iterable[DXFTag]) -> Drawing:
        """Load DXF document from a DXF tag loader, in general an external
        untrusted source.

        Args:
            tag_loader: DXF tag loader

        """
        from .lldxf.tagger import tag_compiler

        tag_loader = tag_compiler(tag_loader)  # type: ignore
        doc = cls()
        doc._load(tag_loader)
        return doc

    @classmethod
    def from_tags(cls, compiled_tags: Iterable[DXFTag]) -> Drawing:
        """Create new drawing from compiled tags. (internal API)"""
        doc = cls()
        doc._load(tagger=compiled_tags)
        return doc

    def _load(self, tagger: Iterable[DXFTag]) -> None:
        # 1st Loading stage: load complete DXF entity structure
        self.is_loading = True
        sections = loader.load_dxf_structure(tagger)
        if "THUMBNAILIMAGE" in sections:
            del sections["THUMBNAILIMAGE"]
        self._load_section_dict(sections)

    def _load_section_dict(self, sections: loader.SectionDict) -> None:
        """Internal API to load a DXF document from a section dict."""
        self.is_loading = True
        # Create header section:
        header_entities: list[Tags] = sections.get("HEADER", [])  # type: ignore
        if header_entities:
            # All header tags are the first DXF structure entity
            self.header = HeaderSection.load(header_entities[0])
        else:
            # Create default header, files without header are by default DXF R12
            self.header = HeaderSection.new(dxfversion=DXF12)
        self._dxfversion = self.header.get("$ACADVER", DXF12)

        # Store original DXF version of loaded file.
        self._loaded_dxfversion = self._dxfversion

        # Content encoding:
        self.encoding = toencoding(self.header.get("$DWGCODEPAGE", "ANSI_1252"))

        # Set handle seed:
        seed: str = self.header.get("$HANDSEED", str(self.entitydb.handles))
        self.entitydb.handles.reset(_validate_handle_seed(seed))

        # Store all necessary DXF entities in the entity database:
        loader.load_and_bind_dxf_content(sections, self)

        # End of 1. loading stage, all entities of the DXF file are
        # stored in the entity database.

        # Create sections:
        self.classes = ClassesSection(
            self, sections.get("CLASSES", None)  # type: ignore
        )
        self.tables = TablesSection(self, sections.get("TABLES", None))  # type: ignore

        # Create *Model_Space and *Paper_Space BLOCK_RECORDS
        # BlockSection setup takes care about the rest:
        self._create_required_block_records()

        # At this point all table entries are required:
        self.blocks = BlocksSection(self, sections.get("BLOCKS", None))  # type: ignore
        self.entities = EntitySection(
            self, sections.get("ENTITIES", None)  # type: ignore
        )
        self.objects = ObjectsSection(
            self, sections.get("OBJECTS", None)  # type: ignore
        )

        # only DXF R2013+
        self.acdsdata = AcDsDataSection(
            self, sections.get("ACDSDATA", None)  # type: ignore
        )

        # Store unmanaged sections as raw tags:
        for name, data in sections.items():
            if name not in const.MANAGED_SECTIONS:
                self.stored_sections.append(StoredSection(data))  # type: ignore

        # Objects section is not initialized!
        self._2nd_loading_stage()

        # DXF version upgrades:
        if self.dxfversion < DXF12:
            logger.info("DXF version upgrade to DXF R12.")
            self.dxfversion = DXF12

        if self.dxfversion == DXF12:
            self.tables.create_table_handles()

        if self.dxfversion in (DXF13, DXF14):
            logger.info("DXF version upgrade to DXF R2000.")
            self.dxfversion = DXF2000
            self.create_all_arrow_blocks()

        # Objects section setup:
        rootdict = self.objects.rootdict
        if rootdict.DXFTYPE != "DICTIONARY":
            raise const.DXFStructureError(
                f"invalid root dictionary entity {str(rootdict)} "
            )
        self.rootdict = rootdict

        # Create missing management tables (DICTIONARY):
        self.objects.setup_object_management_tables(self.rootdict)

        # Setup modelspace- and paperspace layouts:
        self.layouts = Layouts.load(self)

        # Additional work is common to the new and load process:
        self.is_loading = False
        self._finalize_setup()

    def _2nd_loading_stage(self):
        """Load additional resources from entity database into DXF entities.

        e.g. convert handles into DXFEntity() objects

        """
        db = self.entitydb
        for entity in db.values():
            # The post_load_hook() can return a callable, which should be
            # executed, when the DXF document is fully initialized.
            cmd = entity.post_load_hook(self)
            if cmd is not None:
                self._post_init_commands.append(cmd)

    def create_all_arrow_blocks(self):
        """For upgrading DXF R12/13/14 files to R2000, it is necessary to
        create all used arrow blocks before saving the DXF file, else $HANDSEED
        is not the next available handle, which is a problem for AutoCAD.

        Create all known AutoCAD arrows to be on the safe side, because
        references to arrow blocks can be in DIMSTYLE, DIMENSION override,
        LEADER override and maybe other locations.

        """
        from ezdxf.render.arrows import ARROWS

        for arrow_name in ARROWS.__acad__:
            ARROWS.create_block(self.blocks, arrow_name)

    def _create_required_block_records(self):
        if "*Model_Space" not in self.block_records:
            self.block_records.new("*Model_Space")
        if "*Paper_Space" not in self.block_records:
            self.block_records.new("*Paper_Space")

    def saveas(
        self,
        filename: Union[os.PathLike, str],
        encoding: Optional[str] = None,
        fmt: str = "asc",
    ) -> None:
        """Set :class:`Drawing` attribute :attr:`filename` to `filename` and
        write drawing to the file system. Override file encoding by argument
        `encoding`, handle with care, but this option allows you to create DXF
        files for applications that handles file encoding different than
        AutoCAD.

        Args:
            filename: file name as string
            encoding: override default encoding as Python encoding string like ``'utf-8'``
            fmt: ``'asc'`` for ASCII DXF (default) or ``'bin'`` for Binary DXF

        """
        self.filename = str(filename)
        self.save(encoding=encoding, fmt=fmt)

    def save(self, encoding: Optional[str] = None, fmt: str = "asc") -> None:
        """Write drawing to file-system by using the :attr:`filename` attribute
        as filename. Override file encoding by argument `encoding`, handle with
        care, but this option allows you to create DXF files for applications
        that handle file encoding different from AutoCAD.

        Args:
            encoding: override default encoding as Python encoding string like ``'utf-8'``
            fmt: ``'asc'`` for ASCII DXF (default) or ``'bin'`` for Binary DXF

        """
        # DXF R12, R2000, R2004 - ASCII encoding
        # DXF R2007 and newer - UTF-8 encoding
        # in ASCII mode, unknown characters will be escaped as \U+nnnn unicode
        # characters.

        if encoding is None:
            enc = self.output_encoding
        else:
            # override default encoding, for applications that handle encoding
            # different than AutoCAD
            enc = encoding

        if fmt.startswith("asc"):
            fp = io.open(
                self.filename, mode="wt", encoding=enc, errors="dxfreplace"  # type: ignore
            )
        elif fmt.startswith("bin"):
            fp = open(self.filename, "wb")  # type: ignore
        else:
            raise ValueError(f"Unknown output format: '{fmt}'.")
        try:
            self.write(fp, fmt=fmt)  # type: ignore
        finally:
            fp.close()

    def encode(self, s: str) -> bytes:
        """Encode string `s` with correct encoding and error handler."""
        return s.encode(encoding=self.output_encoding, errors="dxfreplace")

    def write(self, stream: Union[TextIO, BinaryIO], fmt: str = "asc") -> None:
        """Write drawing as ASCII DXF to a text stream or as Binary DXF to a
        binary stream. For DXF R2004 (AC1018) and prior open stream with
        drawing :attr:`encoding` and :code:`mode='wt'`. For DXF R2007 (AC1021)
        and later use :code:`encoding='utf-8'`, or better use the later added
        :class:`Drawing` property :attr:`output_encoding` which returns the
        correct encoding automatically. The correct and required error handler
        is :code:`errors='dxfreplace'`!

        If writing to a :class:`StringIO` stream, use :meth:`Drawing.encode` to
        encode the result string from :meth:`StringIO.get_value`::

            binary = doc.encode(stream.get_value())

        Args:
            stream: output text stream or binary stream
            fmt: "asc" for ASCII DXF (default) or "bin" for binary DXF

        """
        # These changes may alter the document content (create new entities, blocks ...)
        # and have to be done before the export and the update of internal structures
        # can be done.
        self.commit_pending_changes()

        dxfversion = self.dxfversion
        if dxfversion == DXF12:
            handles = bool(self.header.get("$HANDLING", 0))
        else:
            handles = True
        if dxfversion > DXF12:
            self.classes.add_required_classes(dxfversion)

        self.update_all()
        if fmt.startswith("asc"):
            tagwriter = TagWriter(
                stream,  # type: ignore
                write_handles=handles,
                dxfversion=dxfversion,
            )
        elif fmt.startswith("bin"):
            tagwriter = BinaryTagWriter(  # type: ignore
                stream,  # type: ignore
                write_handles=handles,
                dxfversion=dxfversion,
                encoding=self.output_encoding,
            )
            tagwriter.write_signature()  # type: ignore
        else:
            raise ValueError(f"Unknown output format: '{fmt}'.")

        self.export_sections(tagwriter)

    def encode_base64(self) -> bytes:
        """Returns DXF document as base64 encoded binary data."""
        stream = io.StringIO()
        self.write(stream)
        # Create binary data:
        binary_data = self.encode(stream.getvalue())
        # Create Windows line endings and do base64 encoding:
        return base64.encodebytes(binary_data.replace(b"\n", b"\r\n"))

    def export_sections(self, tagwriter: AbstractTagWriter) -> None:
        """DXF export sections. (internal API)"""
        dxfversion = tagwriter.dxfversion
        self.header.export_dxf(tagwriter)
        if dxfversion > DXF12:
            self.classes.export_dxf(tagwriter)
        self.tables.export_dxf(tagwriter)
        self.blocks.export_dxf(tagwriter)
        self.entities.export_dxf(tagwriter)
        if dxfversion > DXF12:
            self.objects.export_dxf(tagwriter)
        if self.acdsdata.is_valid:
            self.acdsdata.export_dxf(tagwriter)
        for section in self.stored_sections:
            section.export_dxf(tagwriter)

        tagwriter.write_tag2(0, "EOF")

    def commit_pending_changes(self) -> None:
        self.messenger.broadcast(msgtypes.COMMIT_PENDING_CHANGES)

    def update_all(self) -> None:
        if self.dxfversion > DXF12:
            self.classes.add_required_classes(self.dxfversion)
        self._create_appids()
        self._update_header_vars()
        self.update_extents()
        self.update_limits()
        self._update_metadata()

    def update_extents(self):
        msp = self.modelspace()
        # many applications not maintain these values
        extmin = msp.dxf.extmin
        extmax = msp.dxf.extmax
        if bool(extmin) and bool(extmax):
            self.header["$EXTMIN"] = extmin
            self.header["$EXTMAX"] = extmax
        active_layout = self.active_layout()
        self.header["$PEXTMIN"] = active_layout.dxf.extmin
        self.header["$PEXTMAX"] = active_layout.dxf.extmax

    def update_limits(self):
        msp = self.modelspace()
        self.header["$LIMMIN"] = msp.dxf.limmin
        self.header["$LIMMAX"] = msp.dxf.limmax
        active_layout = self.active_layout()
        self.header["$PLIMMIN"] = active_layout.dxf.limmin
        self.header["$PLIMMAX"] = active_layout.dxf.limmax

    def _update_header_vars(self):
        from ezdxf.lldxf.const import acad_maint_ver

        # set or correct $CMATERIAL handle
        material = self.entitydb.get(self.header.get("$CMATERIAL", None))
        if material is None or material.dxftype() != "MATERIAL":
            handle = "0"
            if "ByLayer" in self.materials:
                handle = self.materials.get("ByLayer").dxf.handle
            elif "Global" in self.materials:
                handle = self.materials.get("Global").dxf.handle
            self.header["$CMATERIAL"] = handle
        # set ACAD maintenance version - same values as used by BricsCAD
        self.header["$ACADMAINTVER"] = acad_maint_ver.get(self.dxfversion, 0)

    def _update_metadata(self):
        self._update_ezdxf_metadata()
        if ezdxf.options.write_fixed_meta_data_for_testing:
            fixed_date = juliandate(datetime(2000, 1, 1, 0, 0))
            self.header["$TDCREATE"] = fixed_date
            self.header["$TDUCREATE"] = fixed_date
            self.header["$TDUPDATE"] = fixed_date
            self.header["$TDUUPDATE"] = fixed_date
            self.header["$VERSIONGUID"] = CONST_GUID
            self.header["$FINGERPRINTGUID"] = CONST_GUID
        else:
            now = datetime.now()
            self.header["$TDUPDATE"] = juliandate(now)
            self.reset_version_guid()
        self.header["$HANDSEED"] = str(self.entitydb.handles)  # next handle
        self.header["$DWGCODEPAGE"] = tocodepage(self.encoding)

    def ezdxf_metadata(self) -> MetaData:
        """Returns the *ezdxf* :class:`ezdxf.document.MetaData` object, which
        manages  *ezdxf* and custom metadata in DXF files.
        For more information see:  :ref:`ezdxf_metadata`.

        """
        return R12MetaData(self) if self.dxfversion <= DXF12 else R2000MetaData(self)

    def _create_ezdxf_metadata(self):
        ezdxf_meta = self.ezdxf_metadata()
        ezdxf_meta[CREATED_BY_EZDXF] = ezdxf_marker_string()

    def _update_ezdxf_metadata(self):
        ezdxf_meta = self.ezdxf_metadata()
        ezdxf_meta[WRITTEN_BY_EZDXF] = ezdxf_marker_string()

    def _create_appid_if_not_exist(self, name: str, flags: int = 0) -> None:
        if name not in self.appids:
            self.appids.new(name, {"flags": flags})

    def _create_appids(self):
        self._create_appid_if_not_exist("HATCHBACKGROUNDCOLOR", 0)
        self._create_appid_if_not_exist("EZDXF", 0)

    @property
    def acad_release(self) -> str:
        """Returns the AutoCAD release abbreviation like "R12" or "R2000"."""
        return const.acad_release.get(self.dxfversion, "unknown")

    @property
    def layers(self) -> LayerTable:
        return self.tables.layers

    @property
    def linetypes(self) -> LinetypeTable:
        return self.tables.linetypes

    @property
    def styles(self) -> TextstyleTable:
        return self.tables.styles

    @property
    def dimstyles(self) -> DimStyleTable:
        return self.tables.dimstyles

    @property
    def ucs(self) -> UCSTable:
        return self.tables.ucs

    @property
    def appids(self) -> AppIDTable:
        return self.tables.appids

    @property
    def views(self) -> ViewTable:
        return self.tables.views

    @property
    def block_records(self) -> BlockRecordTable:
        return self.tables.block_records

    @property
    def viewports(self) -> ViewportTable:
        return self.tables.viewports

    @property
    def plotstyles(self) -> Dictionary:
        return self.rootdict["ACAD_PLOTSTYLENAME"]  # type: ignore

    @property
    def dimension_renderer(self) -> DimensionRenderer:
        return self._dimension_renderer

    @dimension_renderer.setter
    def dimension_renderer(self, renderer: DimensionRenderer) -> None:
        """
        Set your own dimension line renderer if needed.

        see also: ezdxf.render.dimension

        """
        self._dimension_renderer = renderer

    def modelspace(self) -> Modelspace:
        r"""Returns the modelspace layout, displayed as "Model" tab in CAD
        applications, defined by block record named "\*Model_Space".
        """
        return self.layouts.modelspace()

    def layout(self, name: str = "") -> Layout:
        """Returns paperspace layout `name` or the first layout in tab-order
        if no name is given.

        Args:
            name: paperspace name or empty string for the first paperspace in
                tab-order

        Raises:
            KeyError: layout `name` does not exist

        """
        return self.layouts.get(name)

    def paperspace(self, name: str = "") -> Paperspace:
        """Returns paperspace layout `name` or the active paperspace if no
        name is given.

        Args:
            name: paperspace name or empty string for the active paperspace

        Raises:
            KeyError: if the modelspace was acquired or layout `name` does not exist

        """
        if name:
            psp = self.layouts.get(name)
            if isinstance(psp, Paperspace):
                return psp
            raise KeyError("use method modelspace() to acquire the modelspace.")
        else:
            return self.active_layout()

    def active_layout(self) -> Paperspace:
        r"""Returns the active paperspace layout, defined by block record name
        "\*Paper_Space".
        """
        return self.layouts.active_layout()

    def layout_names(self) -> Iterable[str]:
        """Returns all layout names in arbitrary order."""
        return list(self.layouts.names())

    def layout_names_in_taborder(self) -> Iterable[str]:
        """Returns all layout names in tab-order, "Model" is always the first name."""
        return list(self.layouts.names_in_taborder())

    def reset_fingerprint_guid(self):
        """Reset fingerprint GUID."""
        self.header["$FINGERPRINTGUID"] = guid()

    def reset_version_guid(self):
        """Reset version GUID."""
        self.header["$VERSIONGUID"] = guid()

    @property
    def is_acad_compatible(self) -> bool:
        """Returns ``True`` if the current state of the document is compatible to
        AutoCAD.
        """
        return self._acad_compatible

    @property
    def acad_incompatibility_reasons(self) -> list[str]:
        """Returns a list of reasons why this document is not compatible to AutoCAD."""
        return list(self._acad_incompatibility_reason)

    def add_acad_incompatibility_message(self, msg: str):
        """Add a reason why this document is not compatible to AutoCAD.
        (internal API)
        """
        self._acad_compatible = False
        if msg not in self._acad_incompatibility_reason:
            self._acad_incompatibility_reason.add(msg)
            logger.warning(f"DXF document is not compatible to AutoCAD! {msg}.")

    def query(self, query: str = "*") -> EntityQuery:
        """Entity query over all layouts and blocks, excluding the OBJECTS section and
        the resource tables of the TABLES section.

        Args:
            query: query string

        .. seealso::

            :ref:`entity query string` and :ref:`entity queries`

        """
        return EntityQuery(self.chain_layouts_and_blocks(), query)

    def groupby(self, dxfattrib="", key=None) -> dict:
        """Groups DXF entities of all layouts and blocks (excluding the
        OBJECTS section) by a DXF attribute or a key function.

        Args:
            dxfattrib: grouping DXF attribute like "layer"
            key: key function, which accepts a :class:`DXFEntity` as argument
                and returns a hashable grouping key or ``None`` to ignore
                this entity.

        .. seealso::

            :func:`~ezdxf.groupby.groupby` documentation

        """
        return groupby(self.chain_layouts_and_blocks(), dxfattrib, key)

    def chain_layouts_and_blocks(self) -> Iterator[DXFEntity]:
        """Chain entity spaces of all layouts and blocks. Yields an iterator
        for all entities in all layouts and blocks.
        """
        layouts = list(self.layouts_and_blocks())
        return chain.from_iterable(layouts)

    def layouts_and_blocks(self) -> Iterator[GenericLayoutType]:
        """Iterate over all layouts (modelspace and paperspace) and all block definitions."""
        return iter(self.blocks)

    def delete_layout(self, name: str) -> None:
        """Delete paper space layout `name` and all entities owned by this layout.
        Available only for DXF R2000 or later, DXF R12 supports only one
        paperspace, and it can't be deleted.
        """
        if name not in self.layouts:
            raise const.DXFValueError(f"Layout '{name}' does not exist.")
        else:
            self.layouts.delete(name)

    def new_layout(self, name, dxfattribs=None) -> Paperspace:
        """Create a new paperspace layout `name`. Returns a :class:`~ezdxf.layouts.Paperspace`
        object. DXF R12 (AC1009) supports only one paperspace layout, only the active
        paperspace layout is saved, other layouts are dismissed.

        Args:
            name: unique layout name
            dxfattribs: additional DXF attributes for the
                :class:`~ezdxf.entities.layout.DXFLayout` entity

        Raises:
            DXFValueError: paperspace layout `name` already exist

        """
        if name in self.layouts:
            raise const.DXFValueError(f"Layout '{name}' already exists.")
        else:
            return self.layouts.new(name, dxfattribs)

    def page_setup(
        self,
        name: str = "Layout1",
        fmt: str = "ISO A3",
        landscape=True,
    ) -> Paperspace:
        """Creates a new paperspace layout if `name` does not exist or reset the
        existing layout.  This method requires DXF R2000 or newer.
        The paper format name `fmt` defines one of the following paper sizes,
        measures in landscape orientation:

        =========== ======= ======= =======
        Name        Units   Width   Height
        =========== ======= ======= =======
        ISO A0      mm      1189    841
        ISO A1      mm      841     594
        ISO A2      mm      594     420
        ISO A3      mm      420     297
        ISO A4      mm      297     210
        ANSI A      inch    11      8.5
        ANSI B      inch    17      11
        ANSI C      inch    22      17
        ANSI D      inch    34      22
        ANSI E      inch    44      34
        ARCH C      inch    24      18
        ARCH D      inch    36      24
        ARCH E      inch    48      36
        ARCH E1     inch    42      30
        Letter      inch    11      8.5
        Legal       inch    14      8.5
        =========== ======= ======= =======

        The layout uses the associated units of the paper format as drawing
        units, has no margins or offset defined and the scale of the paperspace
        layout is 1:1.

        Args:
            name: paperspace layout name
            fmt: paper format
            landscape: ``True`` for landscape orientation, ``False`` for portrait
                orientation

        """
        from ezdxf.tools.standards import PAGE_SIZES

        if self.acad_release == "R12":
            raise const.DXFVersionError("method call no supported for DXF R12")
        width: float
        height: float
        try:
            units, width, height = PAGE_SIZES[fmt]
        except KeyError:
            raise ValueError(f"unknown paper format: {fmt}")
        if not landscape:
            width, height = height, width
        try:
            psp = self.paperspace(name)
        except KeyError:
            psp = self.new_layout(name)

        psp.page_setup(size=(width, height), margins=(0, 0, 0, 0), units=units)
        return psp

    def acquire_arrow(self, name: str):
        """For standard AutoCAD and ezdxf arrows create block definitions if
        required, otherwise check if block `name` exist. (internal API)

        """
        from ezdxf.render.arrows import ARROWS

        if ARROWS.is_acad_arrow(name) or ARROWS.is_ezdxf_arrow(name):
            ARROWS.create_block(self.blocks, name)
        elif name not in self.blocks:
            raise const.DXFValueError(f'Arrow block "{name}" does not exist.')

    def add_image_def(self, filename: str, size_in_pixel: tuple[int, int], name=None):
        """Add an image definition to the objects section.

        Add an :class:`~ezdxf.entities.image.ImageDef` entity to the drawing
        (objects section). `filename` is the image file name as relative or
        absolute path and `size_in_pixel` is the image size in pixel as (x, y)
        tuple. To avoid dependencies to external packages, `ezdxf` can not
        determine the image size by itself. Returns a
        :class:`~ezdxf.entities.image.ImageDef` entity which is needed to
        create an image reference. `name` is the internal image name, if set to
        ``None``, name is auto-generated.

        Absolute image paths works best for AutoCAD but not perfect, you
        have to update external references manually in AutoCAD, which is not
        possible in TrueView. If the drawing units differ from 1 meter, you
        also have to use: :meth:`set_raster_variables`.

        Args:
            filename: image file name (absolute path works best for AutoCAD)
            size_in_pixel: image size in pixel as (x, y) tuple
            name: image name for internal use, None for using filename as name
                (best for AutoCAD)

        .. seealso::

            :ref:`tut_image`

        """
        if "ACAD_IMAGE_VARS" not in self.rootdict:
            self.objects.set_raster_variables(frame=0, quality=1, units="m")
        if name is None:
            name = filename
        return self.objects.add_image_def(filename, size_in_pixel, name)

    def set_raster_variables(self, frame: int = 0, quality: int = 1, units: str = "m"):
        """Set raster variables.

        Args:
            frame: 0 = do not show image frame; 1 = show image frame
            quality: 0 = draft; 1 = high
            units: units for inserting images. This defines the real world unit
                for one drawing unit for the purpose of inserting and scaling
                images with an associated resolution.

                ===== ===========================
                mm    Millimeter
                cm    Centimeter
                m     Meter (ezdxf default)
                km    Kilometer
                in    Inch
                ft    Foot
                yd    Yard
                mi    Mile
                ===== ===========================

        """
        self.objects.set_raster_variables(frame=frame, quality=quality, units=units)

    def set_wipeout_variables(self, frame=0):
        """Set wipeout variables.

        Args:
            frame: 0 = do not show image frame; 1 = show image frame

        """
        self.objects.set_wipeout_variables(frame=frame)
        var_dict = self.rootdict.get_required_dict("AcDbVariableDictionary")
        var_dict.set_or_add_dict_var("WIPEOUTFRAME", str(frame))

    def add_underlay_def(
        self, filename: str, fmt: str = "ext", name: Optional[str] = None
    ):
        """Add an :class:`~ezdxf.entities.underlay.UnderlayDef` entity to the drawing
        (OBJECTS section). The `filename` is the underlay file name as relative or
        absolute path and `fmt` as string (pdf, dwf, dgn).
        The underlay definition is required to create an underlay reference.

        Args:
            filename: underlay file name
            fmt: file format as string "pdf"|"dwf"|"dgn" or "ext"
                for getting file format from filename extension
            name: pdf format = page number to display; dgn format = "default"; dwf: ????

        .. seealso::

            :ref:`tut_underlay`

        """
        if fmt == "ext":
            fmt = filename[-3:]
        return self.objects.add_underlay_def(filename, fmt, name)

    def add_xref_def(
        self, filename: str, name: str, flags: int = BLK_XREF | BLK_EXTERNAL
    ):
        """Add an external reference (xref) definition to the blocks section.

        Args:
            filename: external reference filename
            name: name of the xref block
            flags: block flags

        """
        self.blocks.new(name=name, dxfattribs={"flags": flags, "xref_path": filename})

    def audit(self) -> Auditor:
        """Checks document integrity and fixes all fixable problems, not
        fixable problems are stored in :attr:`Auditor.errors`.

        If you are messing around with internal structures, call this method
        before saving to be sure to export valid DXF documents, but be aware
        this is a long-running task.

        """
        auditor = Auditor(self)
        auditor.run()
        return auditor

    def validate(self, print_report=True) -> bool:
        """Simple way to run an audit process. Fixes all fixable problems,
        return ``False`` if not fixable errors occurs. Prints a report of resolved and
        unrecoverable errors, if requested.

        Args:
            print_report: print report to stdout

        Returns: ``False`` if unrecoverable errors exist

        """
        auditor = self.audit()
        if print_report:
            auditor.print_fixed_errors()
            auditor.print_error_report()
            if not self.is_acad_compatible:
                print("DXF document is not AutoCAD compatible:")
                for msg in self.acad_incompatibility_reasons:
                    print(msg)
        return len(auditor.errors) == 0

    def set_modelspace_vport(self, height, center=(0, 0), *, dxfattribs=None) -> VPort:
        r"""Set initial view/zoom location for the modelspace, this replaces
        the current "\*Active" viewport configuration
        (:class:`~ezdxf.entities.VPort`) and reset the coordinate system to the
        :ref:`WCS`.

        Args:
             height: modelspace area to view
             center: modelspace location to view in the center of the CAD
                application window.
             dxfattribs: additional DXF attributes for the VPORT entity

        """
        self.viewports.delete_config("*Active")
        dxfattribs = dict(dxfattribs or {})
        vport = cast("VPort", self.viewports.new("*Active", dxfattribs=dxfattribs))
        vport.dxf.center = center
        vport.dxf.height = height
        vport.reset_wcs()
        self.header.reset_wcs()
        return vport


class MetaData(abc.ABC):
    """Manage ezdxf meta-data by dict-like interface. Values are limited to
    strings with a maximum length of 254 characters.
    """

    @abc.abstractmethod
    def __getitem__(self, key: str) -> str:
        """Returns the value for self[`key`].

        Raises:
             KeyError: `key` does not exist
        """
        ...

    def get(self, key: str, default: str = "") -> str:
        """Returns the value for `key`. Returns `default` if `key` not exist."""
        try:
            return self.__getitem__(key)
        except KeyError:
            return safe_string(default, MAX_STR_LEN)

    @abc.abstractmethod
    def __setitem__(self, key: str, value: str) -> None:
        """Set self[`key`] to `value`."""
        ...

    @abc.abstractmethod
    def __delitem__(self, key: str) -> None:
        """Delete self[`key`].

        Raises:
             KeyError: `key` does not exist
        """
        ...

    @abc.abstractmethod
    def __contains__(self, key: str) -> bool:
        """Returns `key` ``in`` self."""
        ...

    def discard(self, key: str) -> None:
        """Remove `key`, does **not** raise an exception if `key` not exist."""
        try:
            self.__delitem__(key)
        except KeyError:
            pass


def ezdxf_marker_string():
    if ezdxf.options.write_fixed_meta_data_for_testing:
        return CONST_MARKER_STRING
    else:
        now = datetime.now(tz=timezone.utc)
        return ezdxf.__version__ + " @ " + now.isoformat()


class R12MetaData(MetaData):
    """Manage ezdxf meta data for DXF version R12 as XDATA of layer "0".

    Layer "0" is mandatory and can not be deleted.

    """

    def __init__(self, doc: Drawing):
        # storing XDATA in layer 0 does not work (Autodesk!)
        self._msp_block = doc.modelspace().block_record.block
        self._data = self._load()

    def __contains__(self, key: str) -> bool:
        return safe_string(key, MAX_STR_LEN) in self._data

    def __getitem__(self, key: str) -> str:
        return self._data[safe_string(key, MAX_STR_LEN)]

    def __setitem__(self, key: str, value: str) -> None:
        self._data[safe_string(key)] = safe_string(value, MAX_STR_LEN)
        self._commit()

    def __delitem__(self, key: str) -> None:
        del self._data[safe_string(key, MAX_STR_LEN)]
        self._commit()

    def _commit(self) -> None:
        # write all metadata as strings with group code 1000
        tags = []
        for key, value in self._data.items():
            tags.append((1000, str(key)))
            tags.append((1000, str(value)))
        self._msp_block.set_xdata("EZDXF", tags)  # type: ignore

    def _load(self) -> dict:
        data = dict()
        if self._msp_block.has_xdata("EZDXF"):  # type: ignore
            xdata = self._msp_block.get_xdata("EZDXF")  # type: ignore
            index = 0
            count = len(xdata) - 1
            while index < count:
                name = xdata[index].value
                data[name] = xdata[index + 1].value
                index += 2
        return data


class R2000MetaData(MetaData):
    """Manage ezdxf metadata for DXF version R2000+ as DICTIONARY object in
    the rootdict.
    """

    def __init__(self, doc: Drawing):
        self._data: Dictionary = doc.rootdict.get_required_dict(
            EZDXF_META, hard_owned=True
        )

    def __contains__(self, key: str) -> bool:
        return safe_string(key, MAX_STR_LEN) in self._data

    def __getitem__(self, key: str) -> str:
        v = self._data[safe_string(key, MAX_STR_LEN)]
        return v.dxf.get("value", "")

    def __setitem__(self, key: str, value: str) -> None:
        self._data.set_or_add_dict_var(
            safe_string(key, MAX_STR_LEN), safe_string(value, MAX_STR_LEN)
        )

    def __delitem__(self, key: str) -> None:
        self._data.remove(safe_string(key, MAX_STR_LEN))


def info(doc: Drawing, verbose=False, content=False, fmt="ASCII") -> list[str]:
    from ezdxf.units import unit_name
    from collections import Counter

    def count(entities, indent="  ") -> Iterator[str]:
        counter: Counter = Counter()
        for e in entities:
            counter[e.dxftype()] += 1
        for name in sorted(counter.keys()):
            yield f"{indent}{name} ({counter[name]})"

    def append_container(table, name: str, container="table"):
        indent = "  "
        data.append(f"{name} {container} entries: {len(table)}")
        if verbose:
            if name == "STYLE":
                names: list[str] = []
                for entry in table:
                    name = entry.dxf.name
                    if name == "":
                        names.append(f'{indent}*shape-file: "{entry.dxf.font}"')
                    else:
                        names.append(f"{indent}{name}")
            else:
                names = [f"{indent}{entry.dxf.name}" for entry in table]
            names.sort()
            data.extend(names)

    def user_vars(kind: str, indent="") -> Iterator[str]:
        for i in range(5):
            name = f"{kind}{i+1}"
            if name in header:
                yield f"{indent}{name}={header[name]}"

    def append_header_var(name: str, indent=""):
        data.append(f"{indent}{name}: {header.get(name, '<undefined>').strip()}")

    header = doc.header
    loaded_dxf_version = doc.loaded_dxfversion
    if loaded_dxf_version is None:
        loaded_dxf_version = doc.dxfversion
    data: list[str] = []
    data.append(f'Filename: "{doc.filename}"')
    data.append(f"Format: {fmt}")
    if loaded_dxf_version != doc.dxfversion:
        msg = f"Loaded content was upgraded from DXF Version {loaded_dxf_version}"
        release = const.acad_release.get(loaded_dxf_version, "")
        if release:
            msg += f" ({release})"
        data.append(msg)
    data.append(f"Release: {doc.acad_release}")
    data.append(f"DXF Version: {doc.dxfversion}")
    if verbose:
        data.append(
            f"Maintenance Version: {header.get('$ACADMAINTVER', '<undefined>')}"
        )
    data.append(f"Codepage: {header.get('$DWGCODEPAGE', 'ANSI_1252')}")
    data.append(f"Encoding: {doc.output_encoding}")
    data.append("Layouts:")
    data.extend([f"  '{name}'" for name in doc.layout_names_in_taborder()])

    measurement = "Metric" if header.get("$MEASUREMENT", 0) else "Imperial"
    if verbose:
        data.append(f"Unit system: {measurement}")
        data.append(f"Modelspace units: {unit_name(doc.units)}")
        append_header_var("$LASTSAVEDBY")
        append_header_var("$HANDSEED")
        append_header_var("$FINGERPRINTGUID")
        append_header_var("$VERSIONGUID")
        data.extend(user_vars(kind="$USERI"))
        data.extend(user_vars(kind="$USERR"))
        for name, value in header.custom_vars:
            data.append(f'Custom property "{name}": "{value}"')

    ezdxf_metadata = doc.ezdxf_metadata()
    if CREATED_BY_EZDXF in ezdxf_metadata:
        data.append(f"Created by ezdxf: {ezdxf_metadata.get(CREATED_BY_EZDXF)}")
    elif verbose:
        data.append("File was not created by ezdxf >= 0.16.4")
    if WRITTEN_BY_EZDXF in ezdxf_metadata:
        data.append(f"Written by ezdxf: {ezdxf_metadata.get(WRITTEN_BY_EZDXF)}")
    elif verbose:
        data.append("File was not written by ezdxf >= 0.16.4")
    if content:
        data.append("Content stats:")
        append_container(doc.layers, "LAYER")
        append_container(doc.linetypes, "LTYPE")
        append_container(doc.styles, "STYLE")
        append_container(doc.dimstyles, "DIMSTYLE")
        append_container(doc.appids, "APPID")
        append_container(doc.ucs, "UCS")
        append_container(doc.views, "VIEW")
        append_container(doc.viewports, "VPORT")
        append_container(doc.block_records, "BLOCK_RECORD")
        if doc.dxfversion > DXF12:
            append_container(list(doc.classes), "CLASS", container="section")
        data.append(f"Entities in modelspace: {len(doc.modelspace())}")
        if verbose:
            data.extend(count(doc.modelspace()))

        data.append(f"Entities in OBJECTS section: {len(doc.objects)}")
        if verbose:
            data.extend(count(doc.objects))

        unknown_entities = _get_unknown_entities(doc)
        if len(unknown_entities):
            data.append(f"Unknown/unsupported entities: {len(unknown_entities)}")
            if verbose:
                data.extend(count(unknown_entities))
    return data


def _get_unknown_entities(doc: Drawing) -> list[DXFEntity]:
    from ezdxf.entities import DXFTagStorage, ACADProxyEntity, OLE2Frame

    data: list[DXFEntity] = []
    for entity in doc.entitydb.values():
        if isinstance(entity, (DXFTagStorage, ACADProxyEntity, OLE2Frame)):
            data.append(entity)
    return data


def custom_export(doc: Drawing, tagwriter: AbstractTagWriter):
    """Export a DXF document via a custom tag writer."""
    dxfversion = doc.dxfversion
    if dxfversion != DXF12:
        tagwriter.write_handles = True
    doc.update_all()
    doc.export_sections(tagwriter)


def export_json_tags(doc: Drawing, compact=True) -> str:
    """Export a DXF document as JSON formatted tags.

    The `compact` format is a list of ``[group-code, value]`` pairs where each pair is
    a DXF tag. The group-code has to be an integer and the value has to be a string,
    integer, float or list of floats for vertices.

    The `verbose` format (`compact` is ``False``) is a list of ``[group-code, value]``
    pairs where each pair is a 1:1 representation of a DXF tag. The group-code has to be
    an integer and the value has to be a string.

    """
    stream = io.StringIO()
    json_writer = JSONTagWriter(stream, dxfversion=doc.dxfversion, compact=compact)
    custom_export(doc, json_writer)
    return stream.getvalue()


def load_json_tags(data: Sequence[Any]) -> Drawing:
    """Load DXF document from JSON formatted tags.

    The expected JSON format is a list of [group-code, value] pairs where each pair is
    a DXF tag. The `compact` and the `verbose` format is supported.

    Args:
        data: JSON data structure as a sequence of [group-code, value] pairs

    """
    from ezdxf.lldxf.tagger import json_tag_loader

    return Drawing.load(json_tag_loader(data))



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import pytest
import os
from io import StringIO
import ezdxf
from ezdxf.lldxf.tagwriter import TagWriter

CIVIL_3D = os.path.join(
    ezdxf.EZDXF_TEST_FILES, "AutodeskProducts/Civil3D_2018.dxf"
)


@pytest.fixture(scope="module")
def doc():
    try:
        # Recover mode is not necessary:
        return ezdxf.readfile(CIVIL_3D)
    except IOError:
        pytest.skip(f'File "{CIVIL_3D}" not found.')


@pytest.mark.parametrize("handle", ["22C1", "22C4", "22C7"])
def test_civil_3d_preserves_binary_data_in_xrecords(doc, handle):
    # Civil 3D stores binary data in XRECORDS as group code 300:
    xrecord = doc.entitydb[handle]
    bin_data = xrecord.tags[1]
    assert bin_data.code == 300
    data = bin_data.value.encode("utf8", errors="surrogateescape")
    assert b"\xed\xb7\x9d" in data, "expected preserved binary data"


BASEDIR = os.path.dirname(__file__)
DATADIR = "data"


@pytest.mark.parametrize(
    ("handle", "n"), [("22C1", 0), ("22C4", 1), ("22C7", 2)]
)
def test_dxf_export_xrecords_with_binary_data(doc, handle, n):
    def expected():
        filename = os.path.join(BASEDIR, DATADIR, f"XRECORD_{n}.bin")
        if not os.path.exists(filename):
            pytest.skip(f"File {filename} not found.")
        with open(filename, "rb") as fp:
            # normalize line endings and remove spaces
            return fp.read().replace(b"\r\n", b"\n").replace(b" ", b"")

    stream = StringIO()
    writer = TagWriter(stream)

    xrecord = doc.entitydb[handle]
    xrecord.export_dxf(writer)

    # remove spaces
    data = doc.encode(stream.getvalue()).replace(b" ", b"")
    assert data == expected()



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
from typing import cast
import pytest
import ezdxf
from ezdxf.entities import Insert, Point, Attrib


def test_mcount_property():
    insert = Insert.new()
    insert.grid(size=(2, 2), spacing=(10, 10))
    assert insert.mcount == 4

    insert.grid(size=(2, 2), spacing=(10, 0))
    assert insert.mcount == 2

    insert.grid(size=(2, 2), spacing=(0, 10))
    assert insert.mcount == 2

    insert.grid(size=(2, 2), spacing=(0, 0))
    assert insert.mcount == 1


class TestSimpleBlock:
    # without ATTRIB, no rotation, no extrusion
    @pytest.fixture(scope="class")
    def doc(self):
        doc = ezdxf.new()
        blk = doc.blocks.new("POINT")
        blk.add_point(location=(0, 0))
        return doc

    @pytest.fixture
    def insert(self, doc):
        msp = doc.modelspace()
        return msp.add_blockref("POINT", (0, 0))

    @pytest.fixture
    def db(self, doc):
        return doc.entitydb

    def test_minsert_normal_spacing(self, insert):
        insert.grid(size=(2, 2), spacing=(10, 10))
        minsert = list(insert.multi_insert())
        assert len(minsert) == 4
        assert minsert[0].dxf.insert == (0, 0)
        assert minsert[1].dxf.insert == (10, 0)
        assert minsert[2].dxf.insert == (0, 10)
        assert minsert[3].dxf.insert == (10, 10)

    def test_discard_minsert_attribs_from_virtual_insert(self, insert):
        insert.grid(size=(2, 2), spacing=(10, 10))
        vinsert = next(insert.multi_insert())
        assert vinsert.dxf.hasattr("row_count") is False
        assert vinsert.dxf.hasattr("column_count") is False
        assert vinsert.dxf.hasattr("row_spacing") is False
        assert vinsert.dxf.hasattr("column_spacing") is False

    def test_minsert_zero_column_spacing(self, insert):
        insert.grid(size=(2, 2), spacing=(10, 0))
        minsert = list(insert.multi_insert())
        assert len(minsert) == 2
        assert minsert[0].dxf.insert == (0, 0)
        assert minsert[1].dxf.insert == (0, 10)

    def test_minsert_zero_row_spacing(self, insert):
        insert.grid(size=(2, 2), spacing=(0, 10))
        minsert = list(insert.multi_insert())
        assert len(minsert) == 2
        assert minsert[0].dxf.insert == (0, 0)
        assert minsert[1].dxf.insert == (10, 0)

    def test_explode(self, insert, db):
        handle = insert.dxf.handle
        insert.grid(size=(2, 2), spacing=(10, 10))
        points = insert.explode()
        db.purge()

        assert insert.is_alive is False
        assert handle not in db
        assert len(points) == 4
        point = cast(Point, points[3])
        assert point.dxf.owner is not None, "not assigned to a layout"
        assert point.get_layout().name == "Model"
        assert point.dxf.location == (10, 10)


class TestInsertAttributes:
    @pytest.fixture(scope="class")
    def doc(self):
        doc = ezdxf.new()
        blk = doc.blocks.new("POINT")
        blk.add_point(location=(0, 0))
        return doc

    @pytest.fixture(scope="class")
    def insert(self, doc):
        msp = doc.modelspace()
        insert = msp.add_blockref("POINT", (0, 0))
        insert.add_attrib("TEST", text="text", insert=(0, 0))
        return insert

    def test_attribs_transformation(self, insert):
        insert.grid(size=(2, 2), spacing=(10, 10))
        attribs = [i.attribs[0] for i in insert.multi_insert()]

        assert len(attribs) == 4
        assert len(set(id(attrib) for attrib in attribs)) == 4
        assert attribs[0].dxf.insert == (0, 0)
        assert attribs[1].dxf.insert == (10, 0)
        assert attribs[2].dxf.insert == (0, 10)
        assert attribs[3].dxf.insert == (10, 10)

    def test_explode(self, insert, doc):
        db = doc.entitydb
        handle = insert.dxf.handle
        insert.grid(size=(2, 2), spacing=(10, 10))
        entities = insert.explode()
        db.purge()

        assert insert.is_alive is False
        assert handle not in db
        assert len(entities) == 8
        # ATTRIB -> TEXT
        attrib = cast(Attrib, entities.query("TEXT")[3])
        assert attrib.dxf.owner is not None, "not assigned to a layout"
        assert attrib.get_layout().name == "Model"
        assert attrib.dxf.insert == (10, 10)


class TestRotatedInsert:
    angle = 90

    @pytest.fixture(scope="class")
    def insert(self):
        doc = ezdxf.new()
        blk = doc.blocks.new("POINT")
        blk.add_point(location=(0, 0))
        msp = doc.modelspace()

        insert = msp.add_blockref("POINT", (0, 0))
        insert.dxf.rotation = self.angle
        # ATTRIB is placed outside of BLOCK in WCS, INSERT rotation is not
        # applied automatically:
        attrib = insert.add_attrib("TEST", text="text", insert=(0, 0))
        attrib.dxf.rotation = self.angle
        return insert

    def test_minsert_transformation(self, insert):
        insert.grid(size=(2, 2), spacing=(10, 10))
        minsert = list(insert.multi_insert())
        assert len(minsert) == 4
        # Rotated 90° counter clockwise:
        assert minsert[0].dxf.insert.isclose((0, 0))
        assert minsert[1].dxf.insert.isclose((0, 10))
        assert minsert[2].dxf.insert.isclose((-10, 0))
        assert minsert[3].dxf.insert.isclose((-10, 10))

    def test_attribs_transformation(self, insert):
        insert.grid(size=(2, 2), spacing=(10, 10))
        attribs = [i.attribs[0] for i in insert.multi_insert()]
        assert len(attribs) == 4
        assert len(set(id(attrib) for attrib in attribs)) == 4
        # Rotated 90° counter clockwise:
        assert attribs[0].dxf.insert.isclose((0, 0))
        assert attribs[1].dxf.insert.isclose((0, 10))
        assert attribs[2].dxf.insert.isclose((-10, 0))
        assert attribs[3].dxf.insert.isclose((-10, 10))


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2011-2021, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.entities import DXFGraphic, DXFTagStorage


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new()
    return doc.modelspace()


def test_delete_polyline3d(msp):
    entity_count = len(msp)
    db = msp.entitydb
    db_count = len(db)
    pline = msp.add_polyline3d([(0, 0, 0), (1, 2, 3), (4, 5, 6)])
    assert entity_count + 1 == len(
        msp
    ), "vertices should be linked to the POLYLINE entity"
    assert db_count + 5 == len(
        msp.entitydb
    ), "database should get 4 vertices and 1 seqend"

    assert pline.seqend.dxf.owner == pline.dxf.handle, "POLYLINE is owner of SEQEND"
    assert pline.seqend.dxf.handle is not None
    assert pline.seqend.dxf.handle in msp.doc.entitydb

    assert len(pline) == 3
    assert pline.vertices[0].dxf.location == (0, 0, 0)
    assert pline.vertices[1].dxf.location == (1, 2, 3)
    assert pline.vertices[2].dxf.location == (4, 5, 6)
    assert pline.get_mode() == "AcDb3dPolyline"
    assert pline.is_3d_polyline is True
    assert pline.is_2d_polyline is False
    assert pline.is_closed is False
    assert list(pline.points()) == [(0, 0, 0), (1, 2, 3), (4, 5, 6)]
    msp.delete_entity(pline)
    assert entity_count == len(msp)

    db.purge()
    assert len(db) == db_count


def test_unlink_unsupported_graphic_entity(msp):
    size = len(msp)
    alien = DXFTagStorage.from_text(ALIEN_ENTITY)
    msp.add_entity(alien)
    assert alien.dxf.handle is not None
    assert alien.dxf.owner is None
    assert len(msp) == size + 1
    msp.unlink_entity(alien)
    assert len(msp) == size
    assert alien.is_alive is True


def test_delete_unsupported_graphic_entity(msp):
    size = len(msp)
    alien = DXFTagStorage.from_text(ALIEN_ENTITY)
    msp.add_entity(alien)
    msp.delete_entity(alien)
    assert len(msp) == size
    assert alien.is_alive is False


def test_cannot_add_invalid_dxf_objects(msp):
    with pytest.raises(ezdxf.DXFTypeError):
        msp.add_entity(DXFTagStorage())


def test_create_line(msp):
    line = msp.add_line((0, 0), (1, 1))
    assert line.dxf.start == (0.0, 0.0)
    assert line.dxf.end == (1.0, 1.0)


def test_create_point(msp):
    point = msp.add_point((1, 2))
    assert point.dxf.location == (1, 2)


def test_create_circle(msp):
    circle = msp.add_circle((3, 3), 5)
    assert circle.dxf.center == (3.0, 3.0)
    assert circle.dxf.radius == 5.0


def test_create_arc(msp):
    arc = msp.add_arc((3, 3), 5, 30, 60)
    assert arc.dxf.center == (3.0, 3.0)
    assert arc.dxf.radius == 5.0
    assert arc.dxf.start_angle == 30.0
    assert arc.dxf.end_angle == 60.0


def test_create_trace(msp):
    trace = msp.add_trace([(0, 0), (1, 0), (1, 1), (0, 1)])
    assert trace[0] == (0, 0)
    assert trace.dxf.vtx1 == (1, 0)
    assert trace[2] == (1, 1)
    assert trace.dxf.vtx3 == (0, 1)


def test_create_solid(msp):
    trace = msp.add_solid([(0, 0), (1, 0), (1, 1)])
    assert trace.dxf.vtx0 == (0, 0)
    assert trace[1] == (1, 0)
    assert trace.dxf.vtx2 == (1, 1)
    assert trace[3] == (1, 1)


def test_create_3dface(msp):
    trace = msp.add_3dface([(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)])
    assert trace.dxf.vtx0 == (0, 0, 0)
    assert trace[1] == (1, 0, 0)
    assert trace.dxf.vtx2 == (1, 1, 0)
    assert trace[3] == (0, 1, 0)


def test_create_text(msp):
    text = msp.add_text("text")
    assert text.dxf.text == "text"


def test_create_shape(msp):
    shape = msp.add_shape("TestShape", (1, 2), 3.0)
    assert shape.dxf.name == "TestShape"
    assert shape.dxf.insert == (1.0, 2.0)
    assert shape.dxf.size == 3
    assert shape.dxf.rotation == 0
    assert shape.dxf.xscale == 1
    assert shape.dxf.oblique == 0


class Checker(DXFGraphic):
    def post_bind_hook(self) -> None:
        self._post_bind_hook = True


def test_post_bind_hook_call(msp):
    checker = Checker()
    checker.doc = msp.doc
    msp.add_entity(checker)
    assert checker._post_bind_hook is True


ALIEN_ENTITY = """0
POINT
100
AcDbEntity
8
0
100
AcDbPoint
10
0.0
20
0.0
30
0.0
"""



from ezdxf.entities import LWPolyline
import pytest


LWPOLYLINE = """  0
LWPOLYLINE
  5
20000
  8
Offline
 62
3 
100
AcDbEntity
100
AcDbPolyline
 70
0
 90
2
 10
1
 20
2
 10
1
 20
3
"""


def test_fix_invalid_located_acdb_entity_group_codes():
    polyline = LWPolyline.from_text(LWPOLYLINE)

    print(str(polyline))
    print(f"Layer: {polyline.dxf.layer}")

    assert polyline.dxf.layer == "Offline"
    assert polyline.dxf.color == 3


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterator, Callable, Any
from .entities import (
    AcisEntity,
    NONE_REF,
    Face,
    Coedge,
    Loop,
    Vertex,
)
from . import sab


class AcisDebugger:
    def __init__(self, root: AcisEntity = NONE_REF, start_id: int = 1):
        self._next_id = start_id - 1
        self._root: AcisEntity = root
        self.entities: dict[int, AcisEntity] = dict()
        if not root.is_none:
            self._store_entities(root)

    @property
    def root(self) -> AcisEntity:
        return self._root

    def _get_id(self) -> int:
        self._next_id += 1
        return self._next_id

    def _store_entities(self, entity: AcisEntity) -> None:
        if not entity.is_none and entity.id == -1:
            entity.id = self._get_id()
        self.entities[entity.id] = entity
        for e in vars(entity).values():
            if isinstance(e, AcisEntity) and e.id == -1:
                self._store_entities(e)

    def set_entities(self, entity: AcisEntity) -> None:
        self.entities.clear()
        self._root = entity
        self._store_entities(entity)

    def walk(self, root: AcisEntity = NONE_REF) -> Iterator[AcisEntity]:
        def _walk(entity: AcisEntity):
            if entity.is_none:
                return
            yield entity
            done.add(entity.id)
            for e in vars(entity).values():
                if isinstance(e, AcisEntity) and e.id not in done:
                    yield from _walk(e)

        if root.is_none:
            root = self._root
        done: set[int] = set()
        yield from _walk(root)

    def filter(
        self, func: Callable[[AcisEntity], bool], entity: AcisEntity = NONE_REF
    ) -> Iterator[Any]:
        if entity.is_none:
            entity = self._root
        yield from filter(func, self.walk(entity))

    def filter_type(
        self, name: str, entity: AcisEntity = NONE_REF
    ) -> Iterator[Any]:
        if entity.is_none:
            entity = self._root
        yield from filter(lambda x: x.type == name, self.walk(entity))

    @staticmethod
    def entity_attributes(entity: AcisEntity, indent: int = 0) -> Iterator[str]:
        indent_str = " " * indent
        for name, data in vars(entity).items():
            if name == "id":
                continue
            yield f"{indent_str}{name}: {data}"

    def face_link_structure(self, face: Face, indent: int = 0) -> Iterator[str]:
        indent_str = " " * indent

        while not face.is_none:
            partner_faces = list(self.partner_faces(face))
            error = ""
            linked_partner_faces = []
            unlinked_partner_faces = []
            for pface_id in partner_faces:
                pface = self.entities.get(pface_id)
                if pface is None:
                    error += f" face {pface_id} does not exist;"
                if isinstance(pface, Face):
                    reverse_faces = self.partner_faces(pface)
                    if face.id in reverse_faces:
                        linked_partner_faces.append(pface_id)
                    else:
                        unlinked_partner_faces.append(pface_id)
                else:
                    error += f" entity {pface_id} is not a face;"
            if unlinked_partner_faces:
                error = f"unlinked partner faces: {unlinked_partner_faces} {error}"
            yield f"{indent_str}{str(face)} >> {partner_faces} {error}"
            face = face.next_face

    @staticmethod
    def partner_faces(face: Face) -> Iterator[int]:
        coedges: list[Coedge] = []
        loop = face.loop
        while not loop.is_none:
            coedges.extend(co for co in loop.coedges())
            loop = loop.next_loop
        for coedge in coedges:
            for partner_coedge in coedge.partner_coedges():
                yield partner_coedge.loop.face.id

    @staticmethod
    def coedge_structure(face: Face, ident: int = 4) -> list[str]:
        lines: list[str] = []
        coedges: list[Coedge] = []
        loop = face.loop

        while not loop.is_none:
            coedges.extend(co for co in loop.coedges())
            loop = loop.next_loop
        for coedge in coedges:
            edge1 = coedge.edge
            sense1 = coedge.sense
            lines.append(f"Coedge={coedge.id} edge={edge1.id} sense={sense1}")
            for partner_coedge in coedge.partner_coedges():
                edge2 = partner_coedge.edge
                sense2 = partner_coedge.sense
                lines.append(
                    f"    Partner Coedge={partner_coedge.id} edge={edge2.id} sense={sense2}"
                )
        ident_str = " " * ident
        return [ident_str + line for line in lines]

    @staticmethod
    def loop_vertices(loop: Loop, indent: int = 0) -> str:
        indent_str = " " * indent
        return f"{indent_str}{loop} >> {list(AcisDebugger.loop_edges(loop))}"

    @staticmethod
    def loop_edges(loop: Loop) -> Iterator[list[int]]:
        coedge = loop.coedge
        first = coedge
        while not coedge.is_none:
            edge = coedge.edge
            sv = edge.start_vertex
            ev = edge.end_vertex
            if coedge.sense:
                yield [ev.id, sv.id]
            else:
                yield [sv.id, ev.id]
            coedge = coedge.next_coedge
            if coedge is first:
                break

    def vertex_to_edge_relation(self) -> Iterator[str]:
        for vertex in (
            e for e in self.entities.values() if isinstance(e, Vertex)
        ):
            edge = vertex.edge
            sv = edge.start_vertex
            ev = edge.end_vertex
            yield f"{vertex}: parent edge is {edge.id}; {sv.id} => {ev.id}; {edge.curve}"

    def is_manifold(self) -> bool:
        for coedge in self.filter_type("coedge"):
            if len(coedge.partner_coedges()) > 1:
                return False
        return True


def dump_sab_as_text(data: bytes) -> Iterator[str]:
    def entity_data(e):
        for tag, value in e:
            name = sab.Tags(tag).name
            yield f"{name} = {value}"

    decoder = sab.Decoder(data)
    header = decoder.read_header()
    yield from header.dumps()
    index = 0
    try:
        for record in decoder.read_records():
            yield f"--------------------- record: {index}"
            yield from entity_data(record)
            index += 1
    except sab.ParsingError as e:
        yield str(e)



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
#
# Basic tools for the LibreCAD Font Format:
# https://github.com/Rallaz/LibreCAD/wiki/lff-definition
from __future__ import annotations
from typing import Sequence, Iterator, Iterable, Optional, no_type_check
from typing_extensions import TypeAlias
from ezdxf.math import Vec2, BoundingBox2d, Matrix44
from ezdxf import path
from .font_measurements import FontMeasurements
from .glyphs import GlyphPath, Glyphs

__all__ = ["loads", "LCFont", "Glyph", "GlyphCache"]


def loads(s: str) -> LCFont:
    lines = s.split("\n")
    name, letter_spacing, word_spacing = parse_properties(lines)
    lcf = LCFont(name, letter_spacing, word_spacing)
    for glyph, parent_code in parse_glyphs(lines):
        lcf.add(glyph, parent_code)
    return lcf


class LCFont:
    """Low level representation of LibreCAD fonts."""
    def __init__(
        self, name: str = "", letter_spacing: float = 0.0, word_spacing: float = 0.0
    ) -> None:
        self.name: str = name
        self.letter_spacing: float = letter_spacing
        self.word_spacing: float = word_spacing
        self._glyphs: dict[int, Glyph] = dict()

    def __len__(self) -> int:
        return len(self._glyphs)

    def __getitem__(self, item: int) -> Glyph:
        return self._glyphs[item]

    def add(self, glyph: Glyph, parent_code: int = 0) -> None:
        if parent_code:
            try:
                parent_glyph = self._glyphs[parent_code]
            except KeyError:
                return
            glyph = parent_glyph.extend(glyph)
        self._glyphs[glyph.code] = glyph

    def get(self, code: int) -> Optional[Glyph]:
        return self._glyphs.get(code, None)


Polyline: TypeAlias = Sequence[Sequence[float]]


class Glyph:
    """Low level representation of a LibreCAD glyph."""
    __slots__ = ("code", "polylines")

    def __init__(self, code: int, polylines: Sequence[Polyline]):
        self.code: int = code
        self.polylines: Sequence[Polyline] = tuple(polylines)

    def extend(self, glyph: Glyph) -> Glyph:
        polylines = list(self.polylines)
        polylines.extend(glyph.polylines)
        return Glyph(glyph.code, polylines)

    def to_path(self) -> GlyphPath:
        from ezdxf.math import OCS

        final_path = path.Path()
        ocs = OCS()
        for polyline in self.polylines:
            p = path.Path()  # empty path is required
            path.add_2d_polyline(
                p, convert_bulge_values(polyline), close=False, elevation=0, ocs=ocs
            )
            final_path.extend_multi_path(p)
        return GlyphPath(final_path)


def convert_bulge_values(polyline: Polyline) -> Iterator[Sequence[float]]:
    # In DXF the bulge value is always stored at the start vertex of the arc.
    last_index = len(polyline) - 1
    for index, vertex in enumerate(polyline):
        bulge = 0.0
        if index < last_index:
            next_vertex = polyline[index + 1]
            try:
                bulge = next_vertex[2]
            except IndexError:
                pass
        yield vertex[0], vertex[1], bulge


def parse_properties(lines: list[str]) -> tuple[str, float, float]:
    font_name = ""
    letter_spacing = 0.0
    word_spacing = 0.0
    for line in lines:
        line = line.strip()
        if not line.startswith("#"):
            continue
        try:
            name, value = line.split(":")
        except ValueError:
            continue

        name = name[1:].strip()
        if name == "Name":
            font_name = value.strip()
        elif name == "LetterSpacing":
            try:
                letter_spacing = float(value)
            except ValueError:
                continue
        elif name == "WordSpacing":
            try:
                word_spacing = float(value)
            except ValueError:
                continue
    return font_name, letter_spacing, word_spacing


def scan_glyphs(lines: Iterable[str]) -> Iterator[list[str]]:
    glyph: list[str] = []
    for line in lines:
        if line.startswith("["):
            if glyph:
                yield glyph
            glyph.clear()
        if line:
            glyph.append(line)
    if glyph:
        yield glyph


def strip_clutter(lines: list[str]) -> Iterator[str]:
    for line in lines:
        line = line.strip()
        if not line.startswith("#"):
            yield line


def scan_int_ex(s: str) -> int:
    from string import hexdigits

    if len(s) == 0:
        return 0
    try:
        end = s.index("]")
    except ValueError:
        end = len(s)
    s = s[1:end].lower()
    s = "".join(c for c in s if c in hexdigits)
    try:
        return int(s, 16)
    except ValueError:
        return 0


def parse_glyphs(lines: list[str]) -> Iterator[tuple[Glyph, int]]:
    code: int
    polylines: list[Polyline] = []
    for glyph in scan_glyphs(strip_clutter(lines)):
        parent_code: int = 0
        polylines.clear()
        line = glyph.pop(0)
        if line[0] != "[":
            continue
        try:
            code = int(line[1 : line.index("]")], 16)
        except ValueError:
            code = scan_int_ex(line)
        if code == 0:
            continue
        line = glyph[0]
        if line.startswith("C"):
            glyph.pop(0)
            try:
                parent_code = int(line[1:], 16)
            except ValueError:
                continue
        polylines = list(parse_polylines(glyph))
        yield Glyph(code, polylines), parent_code


def parse_polylines(lines: Iterable[str]) -> Iterator[Polyline]:
    polyline: list[Sequence[float]] = []
    for line in lines:
        polyline.clear()
        for vertex in line.split(";"):
            values = to_floats(vertex.split(","))
            if len(values) > 1:
                polyline.append(values[:3])
        yield tuple(polyline)


def to_floats(values: Iterable[str]) -> Sequence[float]:
    def strip(value: str) -> float:
        if value.startswith("A"):
            value = value[1:]
        try:
            return float(value)
        except ValueError:
            return 0.0

    return tuple(strip(value) for value in values)


class GlyphCache(Glyphs):
    """Text render engine for LibreCAD fonts with integrated glyph caching."""

    def __init__(self, font: LCFont) -> None:
        self.font: LCFont = font
        self._glyph_cache: dict[int, GlyphPath] = dict()
        self._advance_width_cache: dict[int, float] = dict()
        self.space_width: float = self.font.word_spacing
        self.empty_box: GlyphPath = self.get_empty_box()
        self.font_measurements: FontMeasurements = self._get_font_measurements()

    def get_scaling_factor(self, cap_height: float) -> float:
        try:
            return cap_height / self.font_measurements.cap_height
        except ZeroDivisionError:
            return 1.0

    def get_empty_box(self) -> GlyphPath:
        glyph_A = self.get_shape(65)
        box = BoundingBox2d(glyph_A.control_vertices())
        height = box.size.y
        width = box.size.x
        start = glyph_A.start
        p = path.Path(start)
        p.line_to(start + Vec2(width, 0))
        p.line_to(start + Vec2(width, height))
        p.line_to(start + Vec2(0, height))
        p.close()
        p.move_to(glyph_A.end)
        return GlyphPath(p)

    def _render_shape(self, shape_number) -> GlyphPath:
        try:
            glyph = self.font[shape_number]
        except KeyError:
            if shape_number > 32:
                return self.empty_box
            raise ValueError("space and non-printable characters are not glyphs")
        return glyph.to_path()

    def get_shape(self, shape_number: int) -> GlyphPath:
        if shape_number <= 32:
            raise ValueError("space and non-printable characters are not glyphs")
        try:
            return self._glyph_cache[shape_number].clone()
        except KeyError:
            pass
        glyph = self._render_shape(shape_number)
        self._glyph_cache[shape_number] = glyph
        advance_width = 0.0
        if len(glyph):
            box = glyph.bbox()
            assert box.extmax is not None
            advance_width = box.extmax.x + self.font.letter_spacing
        self._advance_width_cache[shape_number] = advance_width
        return glyph.clone()

    def get_advance_width(self, shape_number: int) -> float:
        if shape_number < 32:
            return 0.0
        if shape_number == 32:
            return self.space_width
        try:
            return self._advance_width_cache[shape_number]
        except KeyError:
            pass
        _ = self.get_shape(shape_number)
        return self._advance_width_cache[shape_number]

    @no_type_check
    def _get_font_measurements(self) -> FontMeasurements:
        # ignore last move_to command, which places the pen at the start of the
        # following glyph
        bbox = BoundingBox2d(self.get_shape(ord("x")).control_vertices())
        baseline = bbox.extmin.y
        x_height = bbox.extmax.y - baseline

        bbox = BoundingBox2d(self.get_shape(ord("A")).control_vertices())
        cap_height = bbox.extmax.y - baseline
        bbox = BoundingBox2d(self.get_shape(ord("p")).control_vertices())
        descender_height = baseline - bbox.extmin.y
        return FontMeasurements(
            baseline=baseline,
            cap_height=cap_height,
            x_height=x_height,
            descender_height=descender_height,
        )

    def get_text_length(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> float:
        scaling_factor = self.get_scaling_factor(cap_height) * width_factor
        return sum(self.get_advance_width(ord(c)) for c in text) * scaling_factor

    def get_text_glyph_paths(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> list[GlyphPath]:
        glyph_paths: list[GlyphPath] = []
        sy = self.get_scaling_factor(cap_height)
        sx = sy * width_factor
        m = Matrix44.scale(sx, sy, 1)
        current_location = 0.0
        for c in text:
            shape_number = ord(c)
            if shape_number > 32:
                glyph = self.get_shape(shape_number)
                m[3, 0] = current_location
                glyph.transform_inplace(m)
                glyph_paths.append(glyph)
            current_location += self.get_advance_width(shape_number) * sx
        return glyph_paths



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
#
# Central import location of frontend dependencies.
# To extract te hpgl2 add-on from the ezdxf package, the following tools have to be
# implemented or extracted too. The dependencies of the implemented backends are not
# listed here.

from ezdxf.math import Vec2, ConstructionCircle, BoundingBox2d, Bezier4P, AnyVec, Matrix44
from ezdxf.path import Path, bbox as path_bbox, transform_paths
from ezdxf.tools.standards import PAGE_SIZES
from ezdxf import colors
NULLVEC2 = Vec2(0, 0)



# Copyright (c) 2020 Manfred Moitzi
# License: MIT License
import pytest
import math
import ezdxf
from ezdxf.entities.lwpolyline import LWPolyline

POINTS = [(0, 0), (1, 0, 1), (2, 0), (3, 0)]


@pytest.fixture
def lwpolyline():
    entity = LWPolyline.new(dxfattribs={"layer": "LAY", "color": 1})
    entity.set_points(POINTS, format="xyb")
    return entity


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new()
    return doc.modelspace()


def test_virtual_entities(lwpolyline):
    result = list(lwpolyline.virtual_entities())
    assert len(result) == 3

    e = result[0]
    assert e.dxftype() == "LINE"
    assert e.dxf.layer == "LAY"
    assert e.dxf.color == 1
    assert e.dxf.start == (0, 0)
    assert e.dxf.end == (1, 0)

    e = result[1]
    assert e.dxftype() == "ARC"
    assert e.dxf.layer == "LAY"
    assert e.dxf.color == 1
    assert e.dxf.center == (1.5, 0)
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 180, abs_tol=1e-12)
    assert math.isclose(e.dxf.end_angle, 0, abs_tol=1e-12)

    assert e.start_point.isclose((1, 0))
    assert e.end_point.isclose((2, 0))

    e = result[2]
    assert e.dxftype() == "LINE"
    assert e.dxf.layer == "LAY"
    assert e.dxf.color == 1
    assert e.dxf.start == (2, 0)
    assert e.dxf.end == (3, 0)


def test_virtual_sub_entities_source_tracking(lwpolyline):
    result = set(e.source_of_copy for e in lwpolyline.virtual_entities())
    assert len(result) == 1, "only one source of copy expected"
    assert lwpolyline in result, "lwpolyline should be the source of copy"


def test_virtual_entities_elevation(lwpolyline):
    lwpolyline = lwpolyline.translate(1, 1, 1)
    assert lwpolyline.dxf.elevation == 1
    result = list(lwpolyline.virtual_entities())
    assert len(result) == 3
    e = result[0]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == (1, 1, 1)
    assert e.dxf.end == (2, 1, 1)

    e = result[1]
    assert e.dxftype() == "ARC"
    assert e.dxf.center == (2.5, 1, 1)
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 180, abs_tol=1e-12)
    assert math.isclose(e.dxf.end_angle, 0, abs_tol=1e-12)

    assert e.start_point.isclose((2, 1, 1))
    assert e.end_point.isclose((3, 1, 1))

    e = result[2]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == (3, 1, 1)
    assert e.dxf.end == (4, 1, 1)


def test_closed_polyline():
    lwpolyline = LWPolyline.new()
    # Create a circle by LWPOLYLINE:
    lwpolyline.set_points([(0, 0, 1), (1, 0, 1)], format="xyb")
    lwpolyline.close(True)

    result = list(lwpolyline.virtual_entities())
    assert len(result) == 2

    e = result[0]
    assert e.dxftype() == "ARC"
    assert e.dxf.center.isclose((0.5, 0))
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 180, abs_tol=1e-12)
    assert math.isclose(e.dxf.end_angle, 0, abs_tol=1e-12)

    e = result[1]
    assert e.dxftype() == "ARC"
    assert e.dxf.center.isclose((0.5, 0))
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 0, abs_tol=1e-12)
    assert math.isclose(abs(e.dxf.end_angle), 180, abs_tol=1e-12)


def test_explode_entities(msp):
    lwpolyline = msp.add_lwpolyline(POINTS, format="xyb")
    assert len(msp) == 1
    result = lwpolyline.explode()
    assert lwpolyline.is_alive is False
    assert len(msp) == 3  # LINE, ARC, LINE
    assert len(result) == 3
    assert msp[-1] is result[2]
    assert msp[-2] is result[1]
    assert msp[-3] is result[0]



#!/usr/bin/env python3
# Copyright (c) 2011-2024 Manfred Moitzi
# License: MIT License
import sys
from setuptools import setup
from setuptools import Extension

# setuptools docs: https://setuptools.pypa.io/en/latest/index.html
# build source distribution
#
#   python setup.py sdist --formats=zip,gztar
#
# build wheels:
#
#   python setup.py bdist_wheel
#
# All Cython modules are optional:
ext_modules = [
    Extension("ezdxf.acc.vector", ["src/ezdxf/acc/vector.pyx"], optional=True),
    Extension("ezdxf.acc.matrix44", ["src/ezdxf/acc/matrix44.pyx"], optional=True),
    Extension("ezdxf.acc.bezier4p", ["src/ezdxf/acc/bezier4p.pyx"], optional=True),
    Extension("ezdxf.acc.bezier3p", ["src/ezdxf/acc/bezier3p.pyx"], optional=True),
    Extension("ezdxf.acc.bspline", ["src/ezdxf/acc/bspline.pyx"], optional=True),
    Extension("ezdxf.acc.construct", ["src/ezdxf/acc/construct.pyx"], optional=True),
    Extension(
        "ezdxf.acc.mapbox_earcut", ["src/ezdxf/acc/mapbox_earcut.pyx"], optional=True
    ),
    Extension("ezdxf.acc.linetypes", ["src/ezdxf/acc/linetypes.pyx"], optional=True),
    Extension("ezdxf.acc.np_support", ["src/ezdxf/acc/np_support.pyx"], optional=True),
]
commands = {}
try:
    from Cython.Distutils import build_ext

    commands = {"build_ext": build_ext}
except ImportError:
    ext_modules = []


PYPY = hasattr(sys, "pypy_version_info")
if PYPY:
    print(
        "C-extensions are disabled for pypy, because JIT compiled Python code "
        "is much faster!"
    )
    ext_modules = []
    commands = {}


def get_version() -> str:
    v = {}
    for line in open("./src/ezdxf/version.py").readlines():
        if line.strip().startswith("__version__"):
            exec(line, v)
            return v["__version__"]
    raise IOError("__version__ string not found")


# static attributes are stored in pyproject.toml
# https://packaging.python.org/en/latest/guides/writing-pyproject-toml/
# https://setuptools.pypa.io/en/latest/index.html
setup(
    version=get_version(),
    cmdclass=commands,
    ext_modules=ext_modules,
)



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.layouts import VirtualLayout
from ezdxf import audit


@pytest.fixture
def layout():
    return VirtualLayout()


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_add_simple_entities(layout):
    layout.add_line(start=(0, 0), end=(1, 0))
    layout.add_circle(center=(0, 0), radius=2)
    layout.add_point(location=(0, 0))
    assert len(layout) == 3


def test_entities_have_no_handle(layout):
    layout.add_point(location=(0, 0))
    entity = layout[0]
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None
    assert entity.doc is None


def test_removing_entiy_does_not_destroy_entity(layout):
    point = layout.add_point(location=(0, 0))
    layout.delete_entity(point)
    assert len(layout) == 0
    assert point.is_alive


def test_purge_destroyed_entities(layout):
    point = layout.add_point((1, 1))
    point.destroy()
    assert len(layout) == 1, "entity should not be removed immediately"
    layout.purge()
    assert (
        len(layout) == 0
    ), "destroyed entity should be removed from entity space"


def test_can_add_entity_to_a_real_layout(layout, doc):
    msp = doc.modelspace()
    point = layout.add_point(location=(0, 0))
    msp.add_entity(point)
    assert point in msp
    assert point.dxf.handle is not None


def test_copy_all_entities_to_a_real_layout(layout, doc):
    msp = doc.modelspace()
    point = layout.add_point(location=(0, 0))
    layout.copy_all_to_layout(msp)
    copy = msp[-1]
    assert len(layout) == 1, "do not remove from virtual layout"
    assert point is not copy, "entity should be copied"
    assert copy.doc is doc, "copy should be assigned to DXF document"
    assert copy.dxf.handle is not None, "copy should have a handle"
    assert (
        copy.dxf.owner == msp.layout_key
    ), "copy should have new layout as owner"
    assert (
        copy.dxf.handle in doc.entitydb
    ), "copy should be stored in document entity database"


def test_move_all_entities_to_a_real_layout(layout, doc):
    msp = doc.modelspace()
    point = layout.add_point(location=(0, 0))
    layout.move_all_to_layout(msp)
    entity = msp[-1]
    assert len(layout) == 0, "remove from virtual layout"
    assert point is entity
    assert entity.doc is doc, "entity should be assigned to DXF document"
    assert entity.dxf.handle is not None, "entity should have a handle"
    assert (
        entity.dxf.owner == msp.layout_key
    ), "entity should have new layout as owner"
    assert (
        entity.dxf.handle in doc.entitydb
    ), "entity should be stored in document entity database"


def test_copy_all_sub_entities_to_a_real_layout(layout, doc):
    msp = doc.modelspace()
    polyline = layout.add_polyline3d([(0, 0), (1, 0), (1, 1)])
    len_db = len(doc.entitydb)
    layout.copy_all_to_layout(msp)
    copy = msp[-1]
    # check copied entity
    assert len(layout) == 1, "do not remove from virtual layout"
    assert polyline is not copy, "POLYLINE  entity should be copied"
    assert copy.doc is doc, "copy should be assigned to the DXF document"
    assert (
        len(doc.entitydb) == len_db + 5
    ), "add 1x POLYLINE, 3x VERTEX and 1x SEQEND to entity database"
    assert copy.dxf.handle is not None, "copy should have a handle"
    assert copy.dxf.owner == msp.layout_key, "copy should have `msp` as owner"
    assert (
        copy.dxf.handle in doc.entitydb
    ), "copy should be stored in document entity database"
    # check VERTEX entities
    vertex = copy.vertices[0]
    assert vertex.doc is doc, "vertices should have a document"
    assert (
        vertex.dxf.handle in doc.entitydb
    ), "vertices should be stored in the entity database"
    # check SEQEND
    assert copy.seqend.doc is doc, "seqend should have a document"
    assert (
        copy.seqend.dxf.handle in doc.entitydb
    ), "seqend should be stored in the entity database"

    # Do any further problems exist?
    auditor = audit.audit(copy, doc)
    assert auditor.has_errors is False and auditor.has_fixes is False


def test_move_all_sub_entities_to_a_real_layout(layout, doc):
    msp = doc.modelspace()
    polyline = layout.add_polyline3d([(0, 0), (1, 0), (1, 1)])
    len_db = len(doc.entitydb)
    layout.move_all_to_layout(msp)
    entity = msp[-1]
    # check moved entity
    assert len(layout) == 0, "remove from virtual layout"
    assert polyline is entity
    assert entity.doc is doc, "entity should be assigned to the DXF document"
    assert (
        len(doc.entitydb) == len_db + 5
    ), "add 1x POLYLINE, 3x VERTEX and 1x SEQEND to entity database"
    assert entity.dxf.handle is not None, "entity should have a handle"
    assert (
        entity.dxf.owner == msp.layout_key
    ), "entity should have `msp` as owner"
    assert (
        entity.dxf.handle in doc.entitydb
    ), "entity should be stored in document entity database"
    # check VERTEX entities
    vertex = entity.vertices[0]
    assert vertex.doc is doc, "vertices should have a document"
    assert (
        vertex.dxf.handle in doc.entitydb
    ), "vertices should be stored in the entity database"
    # check SEQEND
    assert entity.seqend.doc is doc, "seqend should have a document"
    assert (
        entity.seqend.dxf.handle in doc.entitydb
    ), "seqend should be stored in the entity database"

    # Do any further problems exist?
    auditor = audit.audit(entity, doc)
    assert auditor.has_errors is False and auditor.has_fixes is False


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2023, Manfred Moitzi
# License: MIT-License
from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Iterable, Optional
from xml.etree import ElementTree
import math
import re
import logging

from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttributes,
    DefSubclass,
    DXFAttr,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import (
    SUBCLASS_MARKER,
    DXFStructureError,
    DXF2010,
    DXFTypeError,
    DXFValueError,
    InvalidGeoDataException,
)
from ezdxf.lldxf.packedtags import VertexArray
from ezdxf.lldxf.tags import Tags, DXFTag
from ezdxf.math import NULLVEC, Z_AXIS, Y_AXIS, UVec, Vec3, Vec2, Matrix44
from ezdxf.tools.text import split_mtext_string
from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .factory import register_entity
from .copy import default_copy, CopyNotSupported
from .. import units

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = ["GeoData", "MeshVertices"]
logger = logging.getLogger("ezdxf")
acdb_geo_data = DefSubclass(
    "AcDbGeoData",
    {
        # 1 = R2009, but this release has no DXF version,
        # 2 = R2010
        "version": DXFAttr(90, default=2),
        # Handle to host block table record
        "block_record_handle": DXFAttr(330, default="0"),
        # 0 = unknown
        # 1 = local grid
        # 2 = projected grid
        # 3 = geographic (latitude/longitude)
        "coordinate_type": DXFAttr(
            70,
            default=3,
            validator=validator.is_in_integer_range(0, 4),
            fixer=RETURN_DEFAULT,
        ),
        # Design point, reference point in WCS coordinates:
        "design_point": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # Reference point in coordinate system coordinates, valid only when
        # coordinate type is Local Grid:
        "reference_point": DXFAttr(11, xtype=XType.point3d, default=NULLVEC),
        # Horizontal unit scale, factor which converts horizontal design coordinates
        # to meters by multiplication:
        "horizontal_unit_scale": DXFAttr(
            40,
            default=1,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Horizontal units per UnitsValue enumeration. Will be kUnitsUndefined if
        # units specified by horizontal unit scale is not supported by AutoCAD
        # enumeration:
        "horizontal_units": DXFAttr(91, default=1),
        # Vertical unit scale, factor which converts vertical design coordinates
        # to meters by multiplication:
        "vertical_unit_scale": DXFAttr(41, default=1),
        # Vertical units per UnitsValue enumeration. Will be kUnitsUndefined if
        # units specified by vertical unit scale is not supported by AutoCAD
        # enumeration:
        "vertical_units": DXFAttr(92, default=1),
        "up_direction": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # North direction vector (2D)
        # In DXF R2009 the group code 12 is used for source mesh points!
        "north_direction": DXFAttr(
            12,
            xtype=XType.point2d,
            default=Y_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # Scale estimation methods:
        # 1 = None
        # 2 = User specified scale factor;
        # 3 = Grid scale at reference point;
        # 4 = Prismoidal
        "scale_estimation_method": DXFAttr(
            95,
            default=1,
            validator=validator.is_in_integer_range(0, 5),
            fixer=RETURN_DEFAULT,
        ),
        "user_scale_factor": DXFAttr(
            141,
            default=1,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Bool flag specifying whether to do sea level correction:
        "sea_level_correction": DXFAttr(
            294,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "sea_level_elevation": DXFAttr(142, default=0),
        "coordinate_projection_radius": DXFAttr(143, default=0),
        # 303, 303, ..., 301: Coordinate system definition string
        "geo_rss_tag": DXFAttr(302, default="", optional=True),
        "observation_from_tag": DXFAttr(305, default="", optional=True),
        "observation_to_tag": DXFAttr(306, default="", optional=True),
        "observation_coverage_tag": DXFAttr(307, default="", optional=True),
        # 93: Number of Geo-Mesh points
        # mesh definition:
        # source mesh point (12, 22) repeat, mesh_point_count? for version == 1
        # target mesh point (13, 14) repeat, mesh_point_count? for version == 1
        # source mesh point (13, 23) repeat, mesh_point_count? for version > 1
        # target mesh point (14, 24) repeat, mesh_point_count? for version > 1
        # 96:  # Number of faces
        # face index 97 repeat, faces_count
        # face index 98 repeat, faces_count
        # face index 99 repeat, faces_count
    },
)
acdb_geo_data_group_codes = group_code_mapping(acdb_geo_data)
EPSG_3395 = """<?xml version="1.0" encoding="UTF-16" standalone="no" ?>
<Dictionary version="1.0" xmlns="http://www.osgeo.org/mapguide/coordinatesystem">

<ProjectedCoordinateSystem id="WORLD-MERCATOR">
<Name>WORLD-MERCATOR</Name>
<AdditionalInformation>
<ParameterItem type="CsMap">
<Key>CSQuadrantSimplified</Key>
<IntegerValue>1</IntegerValue>
</ParameterItem>
</AdditionalInformation>
<DomainOfValidity>
<Extent>
<GeographicElement>
<GeographicBoundingBox>
<WestBoundLongitude>-180.75</WestBoundLongitude>
<EastBoundLongitude>180.75</EastBoundLongitude>
<SouthBoundLatitude>-80.75</SouthBoundLatitude>
<NorthBoundLatitude>84.75</NorthBoundLatitude>
</GeographicBoundingBox>
</GeographicElement>
</Extent>
</DomainOfValidity>
<DatumId>WGS84</DatumId>
<Axis uom="Meter">
<CoordinateSystemAxis>
<AxisOrder>1</AxisOrder>
<AxisName>Easting</AxisName>
<AxisAbbreviation>E</AxisAbbreviation>
<AxisDirection>East</AxisDirection>
</CoordinateSystemAxis>
<CoordinateSystemAxis>
<AxisOrder>2</AxisOrder>
<AxisName>Northing</AxisName>
<AxisAbbreviation>N</AxisAbbreviation>
<AxisDirection>North</AxisDirection>
</CoordinateSystemAxis>
</Axis>
<Conversion>
<Projection>
<OperationMethodId>Mercator (variant B)</OperationMethodId>
<ParameterValue><OperationParameterId>Longitude of natural origin</OperationParameterId><Value uom="degree">0</Value></ParameterValue>
<ParameterValue><OperationParameterId>Standard Parallel</OperationParameterId><Value uom="degree">0</Value></ParameterValue>
<ParameterValue><OperationParameterId>Scaling factor for coord differences</OperationParameterId><Value uom="unity">1</Value></ParameterValue>
<ParameterValue><OperationParameterId>False easting</OperationParameterId><Value uom="Meter">0</Value></ParameterValue>
<ParameterValue><OperationParameterId>False northing</OperationParameterId><Value uom="Meter">0</Value></ParameterValue>
</Projection>
</Conversion>
</ProjectedCoordinateSystem>
<Alias id="3395" type="CoordinateSystem">
<ObjectId>WORLD-MERCATOR</ObjectId>
<Namespace>EPSG Code</Namespace>
</Alias>

<GeodeticDatum id="WGS84">
<Name>WGS84</Name>
<PrimeMeridianId>Greenwich</PrimeMeridianId>
<EllipsoidId>WGS84</EllipsoidId>
</GeodeticDatum>
<Alias id="6326" type="Datum">
<ObjectId>WGS84</ObjectId>
<Namespace>EPSG Code</Namespace>
</Alias>

<Ellipsoid id="WGS84">
<Name>WGS84</Name>
<SemiMajorAxis uom="meter">6.37814e+06</SemiMajorAxis>
<SecondDefiningParameter>
<SemiMinorAxis uom="meter">6.35675e+06</SemiMinorAxis>
</SecondDefiningParameter>
</Ellipsoid>
<Alias id="7030" type="Ellipsoid">
<ObjectId>WGS84</ObjectId>
<Namespace>EPSG Code</Namespace>
</Alias>

</Dictionary>
"""


class MeshVertices(VertexArray):
    VERTEX_SIZE = 2


def mesh_group_codes(version: int) -> tuple[int, int]:
    return (12, 13) if version < 2 else (13, 14)


@register_entity
class GeoData(DXFObject):
    """DXF GEODATA entity"""

    DXFTYPE = "GEODATA"
    DXFATTRIBS = DXFAttributes(base_class, acdb_geo_data)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2010

    # coordinate_type const
    UNKNOWN = 0
    LOCAL_GRID = 1
    PROJECTED_GRID = 2
    GEOGRAPHIC = 3

    # scale_estimation_method const
    NONE = 1
    USER_SCALE = 2
    GRID_SCALE = 3
    PRISMOIDEAL = 4

    def __init__(self) -> None:
        super().__init__()
        self.source_vertices = MeshVertices()
        self.target_vertices = MeshVertices()
        self.faces: list[Sequence[int]] = []
        self.coordinate_system_definition = ""

    def copy(self, copy_strategy=default_copy):
        raise CopyNotSupported(f"Copying of {self.DXFTYPE} not supported.")

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            version = processor.detect_implementation_version(
                subclass_index=1,
                group_code=90,
                default=2,
            )
            tags = processor.fast_load_dxfattribs(
                dxf, acdb_geo_data_group_codes, 1, log=False
            )
            tags = self.load_coordinate_system_definition(tags)  # type: ignore
            if version > 1:
                self.load_mesh_data(tags, version)
            else:
                # version 1 is not really supported, because the group codes are
                # totally messed up!
                logger.warning(
                    "GEODATA version 1 found, perhaps loaded data is incorrect"
                )
                dxf.discard("north_direction")  # group code issue!!!
        return dxf

    def load_coordinate_system_definition(self, tags: Tags) -> Iterable[DXFTag]:
        # 303, 303, 301, Coordinate system definition string, always XML?
        lines = []
        for tag in tags:
            if tag.code in (301, 303):
                lines.append(tag.value.replace("^J", "\n"))
            else:
                yield tag
        if len(lines):
            self.coordinate_system_definition = "".join(lines)

    def load_mesh_data(self, tags: Iterable[DXFTag], version: int = 2):
        # group codes of version 1 and 2 differ, see DXF reference R2009
        src, target = mesh_group_codes(version)
        face_indices = {97, 98, 99}
        face: list[int] = []
        for code, value in tags:
            if code == src:
                self.source_vertices.append(value)
            elif code == target:
                self.target_vertices.append(value)
            elif code in face_indices:
                if code == 97 and len(face):
                    if len(face) != 3:
                        raise DXFStructureError(
                            f"GEODATA face definition error: invalid index "
                            f"count {len(face)}."
                        )
                    self.faces.append(tuple(face))
                    face.clear()
                face.append(value)
        if face:  # collect last face
            self.faces.append(tuple(face))
        if len(self.source_vertices) != len(self.target_vertices):
            raise DXFStructureError(
                "GEODATA mesh definition error: source and target point count "
                "does not match."
            )

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_geo_data.name)
        if self.dxf.version < 2:
            logger.warning(
                "exporting unsupported GEODATA version 1, this may corrupt "
                "the DXF file!"
            )
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "version",
                "block_record_handle",
                "coordinate_type",
                "design_point",
                "reference_point",
                "horizontal_unit_scale",
                "horizontal_units",
                "vertical_unit_scale",
                "vertical_units",
                "up_direction",
                "north_direction",
                "scale_estimation_method",
                "user_scale_factor",
                "sea_level_correction",
                "sea_level_elevation",
                "coordinate_projection_radius",
            ],
        )
        self.export_coordinate_system_definition(tagwriter)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "geo_rss_tag",
                "observation_from_tag",
                "observation_to_tag",
                "observation_coverage_tag",
            ],
        )
        self.export_mesh_data(tagwriter)

    def export_mesh_data(self, tagwriter: AbstractTagWriter):
        if len(self.source_vertices) != len(self.target_vertices):
            raise DXFStructureError(
                "GEODATA mesh definition error: source and target point count "
                "does not match."
            )
        src, target = mesh_group_codes(self.dxf.version)

        tagwriter.write_tag2(93, len(self.source_vertices))
        for s, t in zip(self.source_vertices, self.target_vertices):
            tagwriter.write_vertex(src, s)
            tagwriter.write_vertex(target, t)

        tagwriter.write_tag2(96, len(self.faces))
        for face in self.faces:
            if len(face) != 3:
                raise DXFStructureError(
                    f"GEODATA face definition error: invalid index "
                    f"count {len(face)}."
                )
            f1, f2, f3 = face
            tagwriter.write_tag2(97, f1)
            tagwriter.write_tag2(98, f2)
            tagwriter.write_tag2(99, f3)

    def export_coordinate_system_definition(self, tagwriter: AbstractTagWriter):
        text = self.coordinate_system_definition.replace("\n", "^J")
        chunks = split_mtext_string(text, size=255)
        if len(chunks) == 0:
            chunks.append("")
        while len(chunks) > 1:
            tagwriter.write_tag2(303, chunks.pop(0))
        tagwriter.write_tag2(301, chunks[0])

    def decoded_units(self) -> tuple[Optional[str], Optional[str]]:
        return units.decode(self.dxf.horizontal_units), units.decode(
            self.dxf.vertical_units
        )

    def get_crs(self) -> tuple[int, bool]:
        """Returns the EPSG index and axis-ordering, axis-ordering is ``True``
        if fist axis is labeled "E" or "W" and ``False`` if first axis is
        labeled "N" or "S".

        If axis-ordering is ``False`` the CRS is not compatible with the
        ``__geo_interface__`` or GeoJSON (see chapter 3.1.1).

        Raises:
            InvalidGeoDataException: for invalid or unknown XML data

        The EPSG number is stored in a tag like:

        .. code::

            <Alias id="27700" type="CoordinateSystem">
              <ObjectId>OSGB1936.NationalGrid</ObjectId>
              <Namespace>EPSG Code</Namespace>
            </Alias>

        The axis-ordering is stored in a tag like:

        .. code::

            <Axis uom="METER">
              <CoordinateSystemAxis>
                <AxisOrder>1</AxisOrder>
                <AxisName>Easting</AxisName>
                <AxisAbbreviation>E</AxisAbbreviation>
                <AxisDirection>east</AxisDirection>
              </CoordinateSystemAxis>
              <CoordinateSystemAxis>
                <AxisOrder>2</AxisOrder>
                <AxisName>Northing</AxisName>
                <AxisAbbreviation>N</AxisAbbreviation>
                <AxisDirection>north</AxisDirection>
              </CoordinateSystemAxis>
            </Axis>

        """
        definition = self.coordinate_system_definition
        try:
            # Remove namespaces so that tags can be searched without prefixing
            # their namespace:
            definition = _remove_xml_namespaces(definition)
            root = ElementTree.fromstring(definition)
        except ElementTree.ParseError:
            raise InvalidGeoDataException(
                "failed to parse coordinate_system_definition as xml"
            )

        crs = None
        for alias in root.findall("Alias"):
            try:
                namespace = alias.find("Namespace").text  # type: ignore
            except AttributeError:
                namespace = ""

            if alias.get("type") == "CoordinateSystem" and namespace == "EPSG Code":
                try:
                    crs = int(alias.get("id"))  # type: ignore
                except ValueError:
                    raise InvalidGeoDataException(
                        f'invalid epsg number: {alias.get("id")}'
                    )
                break

        xy_ordering = None
        for axis in root.findall(".//CoordinateSystemAxis"):
            try:
                axis_order = axis.find("AxisOrder").text  # type: ignore
            except AttributeError:
                axis_order = ""

            if axis_order == "1":
                try:
                    first_axis = axis.find("AxisAbbreviation").text  # type: ignore
                except AttributeError:
                    raise InvalidGeoDataException("first axis not defined")

                if first_axis in ("E", "W"):
                    xy_ordering = True
                elif first_axis in ("N", "S"):
                    xy_ordering = False
                else:
                    raise InvalidGeoDataException(f"unknown first axis: {first_axis}")
                break

        if crs is None:
            raise InvalidGeoDataException("no EPSG code associated with CRS")
        elif xy_ordering is None:
            raise InvalidGeoDataException("could not determine axis ordering")
        else:
            return crs, xy_ordering

    def get_crs_transformation(
        self, *, no_checks: bool = False
    ) -> tuple[Matrix44, int]:
        """Returns the transformation matrix and the EPSG index to transform
        WCS coordinates into CRS coordinates. Because of the lack of proper
        documentation this method works only for tested configurations, set
        argument `no_checks` to ``True`` to use the method for untested geodata
        configurations, but the results may be incorrect.

        Supports only "Local Grid" transformation!

        Raises:
            InvalidGeoDataException: for untested geodata configurations

        """
        epsg, xy_ordering = self.get_crs()

        if not no_checks:
            if (
                self.dxf.coordinate_type != GeoData.LOCAL_GRID
                or self.dxf.scale_estimation_method != GeoData.NONE
                or not math.isclose(self.dxf.user_scale_factor, 1.0)
                or self.dxf.sea_level_correction != 0
                or not math.isclose(self.dxf.sea_level_elevation, 0)
                or self.faces
                or not self.dxf.up_direction.isclose((0, 0, 1))
                or self.dxf.observation_coverage_tag != ""
                or self.dxf.observation_from_tag != ""
                or self.dxf.observation_to_tag != ""
                or not xy_ordering
            ):
                raise InvalidGeoDataException(
                    f"Untested geodata configuration: "
                    f"{self.dxf.all_existing_dxf_attribs()}.\n"
                    f"You can try with no_checks=True but the "
                    f"results may be incorrect."
                )

        source = self.dxf.design_point  # in CAD WCS coordinates
        target = self.dxf.reference_point  # in the CRS of the geodata
        north = self.dxf.north_direction

        # -pi/2 because north is at pi/2 so if the given north is at pi/2, no
        # rotation is necessary:
        theta = -(math.atan2(north.y, north.x) - math.pi / 2)
        transformation = (
            Matrix44.translate(-source.x, -source.y, 0)
            @ Matrix44.scale(
                self.dxf.horizontal_unit_scale, self.dxf.vertical_unit_scale, 1
            )
            @ Matrix44.z_rotate(theta)
            @ Matrix44.translate(target.x, target.y, 0)
        )
        return transformation, epsg

    def setup_local_grid(
        self,
        *,
        design_point: UVec,
        reference_point: UVec,
        north_direction: UVec = (0, 1),
        crs: str = EPSG_3395,
    ) -> None:
        """Setup local grid coordinate system. This method is designed to setup
        CRS similar to `EPSG:3395 World Mercator`, the basic features of the
        CRS should fulfill these assumptions:

            - base unit of reference coordinates is 1 meter
            - right-handed coordinate system: +Y=north/+X=east/+Z=up

        The CRS string is not validated nor interpreted!

        .. hint::

            The reference point must be a 2D cartesian map coordinate and not
            a globe (lon/lat) coordinate like stored in GeoJSON or GPS data.

        Args:
            design_point: WCS coordinates of the CRS reference point
            reference_point: CRS reference point in 2D cartesian coordinates
            north_direction: north direction a 2D vertex, default is (0, 1)
            crs: Coordinate Reference System definition XML string, default is
                the definition string for `EPSG:3395 World Mercator`

        """
        doc = self.doc
        if doc is None:
            raise DXFValueError("Valid DXF document required.")
        wcs_units = doc.units
        if units == 0:
            raise DXFValueError(
                "DXF document requires units to be set, " 'current state is "unitless".'
            )
        meter_factor = units.METER_FACTOR[wcs_units]
        if meter_factor is None:
            raise DXFValueError(f"Unsupported document units: {wcs_units}")
        unit_factor = 1.0 / meter_factor
        # Default settings:
        self.dxf.up_direction = Z_AXIS
        self.dxf.observation_coverage_tag = ""
        self.dxf.observation_from_tag = ""
        self.dxf.observation_to_tag = ""
        self.dxf.scale_estimation_method = GeoData.NONE
        self.dxf.coordinate_type = GeoData.LOCAL_GRID
        self.dxf.sea_level_correction = 0
        self.dxf.horizontal_units = wcs_units
        # Factor from WCS -> CRS (m) e.g. 0.01 for horizontal_units==5 (cm),
        # 1cm = 0.01m
        self.dxf.horizontal_unit_scale = unit_factor
        self.dxf.vertical_units = wcs_units
        self.dxf.vertical_unit_scale = unit_factor

        # User settings:
        self.dxf.design_point = Vec3(design_point)
        self.dxf.reference_point = Vec3(reference_point)
        self.dxf.north_direction = Vec2(north_direction)
        self.coordinate_system_definition = str(crs)


def _remove_xml_namespaces(xml_string: str) -> str:
    return re.sub('xmlns="[^"]*"', "", xml_string)



# Purpose: Import data from another DXF document
# Copyright (c) 2013-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, cast, Union, Optional
import logging
from ezdxf.lldxf import const
from ezdxf.render.arrows import ARROWS
from ezdxf.entities import (
    DXFEntity,
    DXFGraphic,
    Hatch,
    Polyline,
    DimStyle,
    Dimension,
    Viewport,
    Linetype,
    Insert,
)

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.layouts import BaseLayout, Layout

logger = logging.getLogger("ezdxf")

IMPORT_TABLES = ["linetypes", "layers", "styles", "dimstyles"]
IMPORT_ENTITIES = {
    "LINE",
    "POINT",
    "CIRCLE",
    "ARC",
    "TEXT",
    "SOLID",
    "TRACE",
    "3DFACE",
    "SHAPE",
    "POLYLINE",
    "ATTRIB",
    "INSERT",
    "ELLIPSE",
    "MTEXT",
    "LWPOLYLINE",
    "SPLINE",
    "HATCH",
    "MESH",
    "XLINE",
    "RAY",
    "ATTDEF",
    "DIMENSION",
    "LEADER",  # dimension style override not supported!
    "VIEWPORT",
}


class Importer:
    """
    The :class:`Importer` class is central element for importing data from
    other DXF documents.

    Args:
        source: source :class:`~ezdxf.drawing.Drawing`
        target: target :class:`~ezdxf.drawing.Drawing`

    Attributes:
        source: source DXF document
        target: target DXF document
        used_layers: Set of used layer names as string, AutoCAD accepts layer
            names without a LAYER table entry.
        used_linetypes: Set of used linetype names as string, these linetypes
            require a TABLE entry or AutoCAD will crash.
        used_styles: Set of used text style names, these text styles require
            a TABLE entry or AutoCAD will crash.
        used_dimstyles: Set of used dimension style names, these dimension
            styles require a TABLE entry or AutoCAD will crash.

    """

    def __init__(self, source: Drawing, target: Drawing):
        self.source: Drawing = source
        self.target: Drawing = target

        self.used_layers: set[str] = set()
        self.used_linetypes: set[str] = set()
        self.used_styles: set[str] = set()
        self.used_shape_files: set[str] = set()  # style entry without a name!
        self.used_dimstyles: set[str] = set()
        self.used_arrows: set[str] = set()
        self.handle_mapping: dict[str, str] = dict()  # old_handle: new_handle

        # collects all imported INSERT entities, for later name resolving.
        self.imported_inserts: list[DXFEntity] = list()  # imported inserts

        # collects all imported block names and their assigned new name
        # imported_block[original_name] = new_name
        self.imported_blocks: dict[str, str] = dict()
        self._default_plotstyle_handle = target.plotstyles["Normal"].dxf.handle
        self._default_material_handle = target.materials["Global"].dxf.handle

    def _add_used_resources(self, entity: DXFEntity) -> None:
        """Register used resources."""
        self.used_layers.add(entity.get_dxf_attrib("layer", "0"))
        self.used_linetypes.add(entity.get_dxf_attrib("linetype", "BYLAYER"))
        if entity.is_supported_dxf_attrib("style"):
            self.used_styles.add(entity.get_dxf_attrib("style", "Standard"))
        if entity.is_supported_dxf_attrib("dimstyle"):
            self.used_dimstyles.add(entity.get_dxf_attrib("dimstyle", "Standard"))

    def _add_dimstyle_resources(self, dimstyle: DimStyle) -> None:
        self.used_styles.add(dimstyle.get_dxf_attrib("dimtxsty", "Standard"))
        self.used_linetypes.add(dimstyle.get_dxf_attrib("dimltype", "BYLAYER"))
        self.used_linetypes.add(dimstyle.get_dxf_attrib("dimltex1", "BYLAYER"))
        self.used_linetypes.add(dimstyle.get_dxf_attrib("dimltex2", "BYLAYER"))
        self.used_arrows.add(dimstyle.get_dxf_attrib("dimblk", ""))
        self.used_arrows.add(dimstyle.get_dxf_attrib("dimblk1", ""))
        self.used_arrows.add(dimstyle.get_dxf_attrib("dimblk2", ""))
        self.used_arrows.add(dimstyle.get_dxf_attrib("dimldrblk", ""))

    def _add_linetype_resources(self, linetype: Linetype) -> None:
        if not linetype.pattern_tags.is_complex_type():
            return
        style_handle = linetype.pattern_tags.get_style_handle()
        style = self.source.entitydb.get(style_handle)
        if style is None:
            return
        if style.dxf.name == "":
            # Shape file entries have no name!
            self.used_shape_files.add(style.dxf.font)
        else:
            self.used_styles.add(style.dxf.name)

    def import_tables(
        self, table_names: Union[str, Iterable[str]] = "*", replace=False
    ) -> None:
        """Import DXF tables from the source document into the target document.

        Args:
            table_names: iterable of tables names as strings, or a single table
                name as string or "*" for all supported tables
            replace: ``True`` to replace already existing table entries else
                ignore existing entries

        Raises:
            TypeError: unsupported table type

        """
        if isinstance(table_names, str):
            if table_names == "*":  # import all supported tables
                table_names = IMPORT_TABLES
            else:  # import one specific table
                table_names = (table_names,)
        for table_name in table_names:
            self.import_table(table_name, entries="*", replace=replace)

    def import_table(
        self, name: str, entries: Union[str, Iterable[str]] = "*", replace=False
    ) -> None:
        """
        Import specific table entries from the source document into the
        target document.

        Args:
            name: valid table names are "layers", "linetypes" and "styles"
            entries: Iterable of table names as strings, or a single table name
                or "*" for all table entries
            replace: ``True`` to replace the already existing table entry else
                ignore existing entries

        Raises:
            TypeError: unsupported table type

        """
        if name not in IMPORT_TABLES:
            raise TypeError(f'Table "{name}" import not supported.')
        source_table = getattr(self.source.tables, name)
        target_table = getattr(self.target.tables, name)

        if isinstance(entries, str):
            if entries == "*":  # import all table entries
                entries = (entry.dxf.name for entry in source_table)
            else:  # import just one table entry
                entries = (entries,)
        for entry_name in entries:
            try:
                table_entry = source_table.get(entry_name)
            except const.DXFTableEntryError:
                logger.warning(
                    f'Required table entry "{entry_name}" in table f{name} '
                    f"not found."
                )
                continue
            entry_name = table_entry.dxf.name
            if entry_name in target_table:
                if replace:
                    logger.debug(
                        f'Replacing already existing entry "{entry_name}" '
                        f"of {name} table."
                    )
                    target_table.remove(table_entry.dxf.name)
                else:
                    logger.debug(
                        f'Discarding already existing entry "{entry_name}" '
                        f"of {name} table."
                    )
                    continue

            if name == "layers":
                self.used_linetypes.add(
                    table_entry.get_dxf_attrib("linetype", "Continuous")
                )
            elif name == "dimstyles":
                self._add_dimstyle_resources(table_entry)
            elif name == "linetypes":
                self._add_linetype_resources(table_entry)

            # Duplicate table entry:
            new_table_entry = self._duplicate_table_entry(table_entry)
            target_table.add_entry(new_table_entry)

            # Register resource handles for mapping:
            self.handle_mapping[table_entry.dxf.handle] = new_table_entry.dxf.handle

    def import_shape_files(self, fonts: set[str]) -> None:
        """Import shape file table entries from the source document into the
        target document.
        Shape file entries are stored in the styles table but without a name.

        """
        for font in fonts:
            table_entry = self.source.styles.find_shx(font)
            # copy is not necessary, just create a new entry:
            new_table_entry = self.target.styles.get_shx(font)
            if table_entry:
                # Register resource handles for mapping:
                self.handle_mapping[table_entry.dxf.handle] = new_table_entry.dxf.handle
            else:
                logger.warning(f'Required shape file entry "{font}" not found.')

    def _set_table_entry_dxf_attribs(self, entity: DXFEntity) -> None:
        entity.doc = self.target
        if entity.dxf.hasattr("plotstyle_handle"):
            entity.dxf.plotstyle_handle = self._default_plotstyle_handle
        if entity.dxf.hasattr("material_handle"):
            entity.dxf.material_handle = self._default_material_handle

    def _duplicate_table_entry(self, entry: DXFEntity) -> DXFEntity:
        # duplicate table entry
        new_entry = new_clean_entity(entry)
        self._set_table_entry_dxf_attribs(entry)

        # create a new handle and add entity to target entity database
        self.target.entitydb.add(new_entry)
        return new_entry

    def import_entity(
        self, entity: DXFEntity, target_layout: Optional[BaseLayout] = None
    ) -> None:
        """
        Imports a single DXF `entity` into `target_layout` or the modelspace
        of the target document, if `target_layout` is ``None``.

        Args:
            entity: DXF entity to import
            target_layout: any layout (modelspace, paperspace or block) from
                the target document

        Raises:
            DXFStructureError: `target_layout` is not a layout of target document

        """

        def set_dxf_attribs(e):
            e.doc = self.target
            # remove invalid resources
            e.dxf.discard("plotstyle_handle")
            e.dxf.discard("material_handle")
            e.dxf.discard("visualstyle_handle")

        if target_layout is None:
            target_layout = self.target.modelspace()
        elif target_layout.doc != self.target:
            raise const.DXFStructureError(
                "Target layout has to be a layout or block from the target " "document."
            )

        dxftype = entity.dxftype()
        if dxftype not in IMPORT_ENTITIES:
            logger.debug(f"Import of {str(entity)} not supported")
            return
        self._add_used_resources(entity)

        try:
            new_entity = cast(DXFGraphic, new_clean_entity(entity))
        except const.DXFTypeError:
            logger.debug(f"Copying for DXF type {dxftype} not supported.")
            return

        set_dxf_attribs(new_entity)
        self.target.entitydb.add(new_entity)
        target_layout.add_entity(new_entity)

        try:  # additional processing
            getattr(self, "_import_" + dxftype.lower())(new_entity)
        except AttributeError:
            pass

    def _import_insert(self, insert: Insert):
        self.imported_inserts.append(insert)
        # remove all possible source document dependencies from sub entities
        for attrib in insert.attribs:
            remove_dependencies(attrib)

    def _import_polyline(self, polyline: Polyline):
        # remove all possible source document dependencies from sub entities
        for vertex in polyline.vertices:
            remove_dependencies(vertex)

    def _import_hatch(self, hatch: Hatch):
        hatch.dxf.discard("associative")

    def _import_viewport(self, viewport: Viewport):
        viewport.dxf.discard("sun_handle")
        viewport.dxf.discard("clipping_boundary_handle")
        viewport.dxf.discard("ucs_handle")
        viewport.dxf.discard("ucs_base_handle")
        viewport.dxf.discard("background_handle")
        viewport.dxf.discard("shade_plot_handle")
        viewport.dxf.discard("visual_style_handle")
        viewport.dxf.discard("ref_vp_object_1")
        viewport.dxf.discard("ref_vp_object_2")
        viewport.dxf.discard("ref_vp_object_3")
        viewport.dxf.discard("ref_vp_object_4")

    def _import_dimension(self, dimension: Dimension):
        if dimension.virtual_block_content:
            for entity in dimension.virtual_block_content:
                if isinstance(entity, Insert):  # import arrow blocks
                    self.import_block(entity.dxf.name, rename=False)
                self._add_used_resources(entity)
        else:
            logger.error("The required geometry block for DIMENSION is not defined.")

    def import_entities(
        self,
        entities: Iterable[DXFEntity],
        target_layout: Optional[BaseLayout] = None,
    ) -> None:
        """Import all `entities` into `target_layout` or the modelspace of the
        target document, if `target_layout` is ``None``.

        Args:
            entities: Iterable of DXF entities
            target_layout: any layout (modelspace, paperspace or block) from
                the target document

        Raises:
            DXFStructureError: `target_layout` is not a layout of target document

        """
        for entity in entities:
            self.import_entity(entity, target_layout)

    def import_modelspace(self, target_layout: Optional[BaseLayout] = None) -> None:
        """Import all entities from source modelspace into `target_layout` or
        the modelspace of the target document, if `target_layout` is ``None``.

        Args:
            target_layout: any layout (modelspace, paperspace or block) from
                the target document

        Raises:
            DXFStructureError: `target_layout` is not a layout of target document

        """
        self.import_entities(self.source.modelspace(), target_layout=target_layout)

    def recreate_source_layout(self, name: str) -> Layout:
        """Recreate source paperspace layout `name` in the target document.
        The layout will be renamed if `name` already exist in the target
        document. Returns target modelspace for layout name "Model".

        Args:
            name: layout name as string

        Raises:
            KeyError: if source layout `name` not exist

        """

        def get_target_name():
            tname = name
            base_name = name
            count = 1
            while tname in self.target.layouts:
                tname = base_name + str(count)
                count += 1

            return tname

        def clear(dxfattribs: dict) -> dict:
            def discard(name: str):
                try:
                    del dxfattribs[name]
                except KeyError:
                    pass

            discard("handle")
            discard("owner")
            discard("taborder")
            discard("shade_plot_handle")
            discard("block_record_handle")
            discard("viewport_handle")
            discard("ucs_handle")
            discard("base_ucs_handle")
            return dxfattribs

        if name.lower() == "model":
            return self.target.modelspace()

        source_layout = self.source.layouts.get(name)  # raises KeyError
        target_name = get_target_name()
        dxfattribs = clear(source_layout.dxf_layout.dxfattribs())
        target_layout = self.target.layouts.new(target_name, dxfattribs=dxfattribs)
        return target_layout

    def import_paperspace_layout(self, name: str) -> Layout:
        """Import paperspace layout `name` into the target document.

        Recreates the source paperspace layout in the target document, renames
        the target paperspace if a paperspace with same `name` already exist
        and imports all entities from the source paperspace into the target
        paperspace.

        Args:
            name: source paper space name as string

        Returns: new created target paperspace :class:`Layout`

        Raises:
            KeyError: source paperspace does not exist
            DXFTypeError: invalid modelspace import

        """
        if name.lower() == "model":
            raise const.DXFTypeError(
                "Can not import modelspace, use method import_modelspace()."
            )
        source_layout = self.source.layouts.get(name)
        target_layout = self.recreate_source_layout(name)
        self.import_entities(source_layout, target_layout)
        return target_layout

    def import_paperspace_layouts(self) -> None:
        """Import all paperspace layouts and their content into the target
        document.
        Target layouts will be renamed if a layout with the same name already
        exist. Layouts will be imported in original tab order.

        """
        for name in self.source.layouts.names_in_taborder():
            if name.lower() != "model":  # do not import modelspace
                self.import_paperspace_layout(name)

    def import_blocks(self, block_names: Iterable[str], rename=False) -> None:
        """Import all BLOCK definitions from source document.

        If a BLOCK already exist the BLOCK will be renamed if argument
        `rename` is ``True``, otherwise the existing BLOCK in the target
        document will be used instead of the BLOCK from the source document.
        Required name resolving for imported BLOCK references (INSERT), will be
        done in the :meth:`Importer.finalize` method.

        Args:
            block_names: names of BLOCK definitions to import
            rename: rename BLOCK if a BLOCK with the same name already exist in
                target document

        Raises:
            ValueError: BLOCK in source document not found (defined)

        """
        for block_name in block_names:
            self.import_block(block_name, rename=rename)

    def import_block(self, block_name: str, rename=True) -> str:
        """Import one BLOCK definition from source document.

        If the BLOCK already exist the BLOCK will be renamed if argument
        `rename` is ``True``, otherwise the existing BLOCK in the target
        document will be used instead of the BLOCK in the source document.
        Required name resolving for imported block references (INSERT), will be
        done in the :meth:`Importer.finalize` method.

        To replace an existing BLOCK in the target document, just delete it
        before importing data:
        :code:`target.blocks.delete_block(block_name, safe=False)`

        Args:
            block_name: name of BLOCK to import
            rename: rename BLOCK if a BLOCK with the same name already exist in
                target document

        Returns: (renamed) BLOCK name

        Raises:
            ValueError: BLOCK in source document not found (defined)

        """

        def get_new_block_name() -> str:
            num = 0
            name = block_name
            while name in target_blocks:
                name = block_name + str(num)
                num += 1
            return name

        try:  # already imported block?
            return self.imported_blocks[block_name]
        except KeyError:
            pass

        try:
            source_block = self.source.blocks[block_name]
        except const.DXFKeyError:
            raise ValueError(f'Source block "{block_name}" not found.')

        target_blocks = self.target.blocks
        if (block_name in target_blocks) and (rename is False):
            self.imported_blocks[block_name] = block_name
            return block_name

        new_block_name = get_new_block_name()
        block = source_block.block
        assert block is not None
        target_block = target_blocks.new(
            new_block_name,
            base_point=block.dxf.base_point,
            dxfattribs={
                "description": block.dxf.description,
                "flags": block.dxf.flags,
                "xref_path": block.dxf.xref_path,
            },
        )
        self.import_entities(source_block, target_layout=target_block)
        self.imported_blocks[block_name] = new_block_name
        return new_block_name

    def _create_missing_arrows(self):
        """Create or import required arrow blocks, used by the LEADER or the
        DIMSTYLE entity, which are not imported automatically because they are
        not used in an anonymous DIMENSION geometry BLOCK.

        """
        self.used_arrows.discard(
            ""
        )  # standard default arrow '' needs no block definition
        for arrow_name in self.used_arrows:
            if ARROWS.is_acad_arrow(arrow_name):
                self.target.acquire_arrow(arrow_name)
            else:
                self.import_block(arrow_name, rename=False)

    def _resolve_inserts(self) -> None:
        """Resolve BLOCK names of imported BLOCK reference entities (INSERT).

        This is required for the case the name of the imported BLOCK collides
        with an already existing BLOCK in the target document and the conflict
        resolving method is "rename".

        """
        while len(self.imported_inserts):
            inserts = list(self.imported_inserts)
            # clear imported inserts, block import may append additional inserts
            self.imported_inserts = []
            for insert in inserts:
                block_name = self.import_block(insert.dxf.name)
                insert.dxf.name = block_name

    def _import_required_table_entries(self) -> None:
        """Import required table entries collected while importing entities
        into the target document.

        """
        # 1. dimstyles import adds additional required linetype and style
        # resources and required arrows
        if len(self.used_dimstyles):
            self.import_table("dimstyles", self.used_dimstyles)

        # 2. layers import adds additional required linetype resources
        if len(self.used_layers):
            self.import_table("layers", self.used_layers)

        # 3. complex linetypes adds additional required style resources
        if len(self.used_linetypes):
            self.import_table("linetypes", self.used_linetypes)

        # 4. Text styles do not add additional required resources
        if len(self.used_styles):
            self.import_table("styles", self.used_styles)

        # 5. Shape files are text style entries without a name
        if len(self.used_shape_files):
            self.import_shape_files(self.used_shape_files)

        # 6. Update text style handles of imported complex linetypes:
        self.update_complex_linetypes()

    def _add_required_complex_linetype_resources(self):
        for ltype_name in self.used_linetypes:
            try:
                ltype = self.source.linetypes.get(ltype_name)
            except const.DXFTableEntryError:
                continue
            self._add_linetype_resources(ltype)

    def update_complex_linetypes(self):
        std_handle = self.target.styles.get("STANDARD").dxf.handle
        for linetype in self.target.linetypes:
            if linetype.pattern_tags.is_complex_type():
                old_handle = linetype.pattern_tags.get_style_handle()
                new_handle = self.handle_mapping.get(old_handle, std_handle)
                linetype.pattern_tags.set_style_handle(new_handle)

    def finalize(self) -> None:
        """Finalize the import by importing required table entries and BLOCK
        definitions, without finalization the target document is maybe invalid
        for AutoCAD. Call the :meth:`~Importer.finalize()` method as last step
        of the import process.

        """
        self._resolve_inserts()
        self._add_required_complex_linetype_resources()
        self._import_required_table_entries()
        self._create_missing_arrows()


def new_clean_entity(entity: DXFEntity, keep_xdata: bool = False) -> DXFEntity:
    """Copy entity and remove all external dependencies.

    Args:
        entity: DXF entity
        keep_xdata: keep xdata flag

    """
    new_entity = entity.copy()
    new_entity.doc = None
    return remove_dependencies(new_entity, keep_xdata=keep_xdata)


def remove_dependencies(entity: DXFEntity, keep_xdata: bool = False) -> DXFEntity:
    """Remove all external dependencies.

    Args:
        entity: DXF entity
        keep_xdata: keep xdata flag

    """
    entity.appdata = None
    entity.reactors = None
    entity.extension_dict = None
    if not keep_xdata:
        entity.xdata = None
    return entity
